"""RNA visualization and graphical analysis toolset
RNA vi                g         a        t    e

A Jupyter-compatible toolset for the exploration of RNA structure and chemical
probing data.
"""

from rnavigate.rnavigate import Sample
from rnavigate.plotting_functions import (
    plot_options,
    plot_alignment,
    plot_arcs,
    plot_arcs_compare,
    plot_circle,
    plot_disthist,
    plot_heatmap,
    plot_linreg,
    plot_mol,
    plot_ntdist,
    plot_profile,
    plot_qc,
    plot_roc,
    plot_shapemapper,
    plot_skyline,
    plot_ss,
)

from rnavigate import data
from rnavigate import plots
from rnavigate import styles

__version__ = "1.0.0"
__author__ = "Patrick S. Irving"
__email__ = "psirving@email.unc.edu"

__all__ = [
    "Sample",
    # plotting functions
    "plot_options",
    "plot_alignment",
    "plot_arcs",
    "plot_arcs_compare",
    "plot_circle",
    "plot_disthist",
    "plot_heatmap",
    "plot_linreg",
    "plot_mol",
    "plot_ntdist",
    "plot_profile",
    "plot_qc",
    "plot_roc",
    "plot_shapemapper",
    "plot_skyline",
    "plot_ss",
    # modules
    "data",
    "plots",
    "styles",
]



import numpy as np
import matplotlib.cm as cm
import matplotlib.colors as mpc


class ScalarMappable(cm.ScalarMappable):
    """Used to map scalar values to a color and to create a colorbar plot.

    Parameters
    ----------
    cmap : str, tuple, float, or list
        A valid mpl color, list of valid colors or a valid colormap name
    normalization : "min_max", "0_1", "none", or "bins"
        The type of normalization to use when mapping values to colors
    values : list
        The values to use when normalizing the data
    title : str, defaults to ""
        The title of the colorbar.
    tick_labels : list, defaults to None
        The labels to use for the colorbar ticks. If None, values are
        determined automatically.
    **cbar_args : dict
        Additional arguments to pass to the colorbar function

    Attributes
    ----------
    rnav_norm : str
        The type of normalization to use when mapping values to colors
    rnav_vals : list
        The values to use when normalizing the data
    rnav_cmap : list
        The colors to use when mapping values to colors
    cbar_args : dict
        Additional arguments to pass to the colorbar function
    tick_labels : list
        The labels to use for the colorbar ticks. If None, values are
        determined automatically.
    title : str
        The title of the colorbar.
    """

    def __init__(
        self, cmap, normalization, values, title="", tick_labels=None, **cbar_args
    ):
        """Initialize the ScalarMappable object"""
        cmap = self.get_cmap(cmap)
        norm = self.get_norm(normalization, values, cmap)
        super().__init__(norm, cmap)
        self.rnav_norm = normalization
        self.rnav_vals = values
        self.rnav_cmap = cmap(np.arange(cmap.N))
        self.cbar_args = cbar_args
        self.tick_labels = tick_labels
        self.title = title

    def is_equivalent_to(self, cmap2):
        """Check if two ScalarMappable objects are equivalent.

        Parameters
        ----------
        cmap2 : ScalarMappable
            The ScalarMappable object to compare to

        Returns
        -------
        bool
            True if the two ScalarMappable objects are equivalent, False
            otherwise
        """
        return (
            (self.rnav_norm == cmap2.rnav_norm)
            and (self.rnav_vals == cmap2.rnav_vals)
            and (len(self.rnav_cmap) == len(cmap2.rnav_cmap))
            and mpc.same_color(self.rnav_cmap, cmap2.rnav_cmap)
            and (self.tick_labels == cmap2.tick_labels)
            and (self.cbar_args == cmap2.cbar_args)
            and (self.title == cmap2.title)
        )

    def values_to_hexcolors(self, values, alpha=1.0):
        """Map values to colors and return a list of hex colors.

        Parameters
        ----------
        values : list
            The values to map to colors
        alpha : float, defaults to 1.0
            The alpha value to use for the colors

        Returns
        -------
        list of strings
            A list of hex colors
        """
        colors = super().to_rgba(x=values, alpha=alpha)
        return np.array([mpc.to_hex(c, keep_alpha=True) for c in colors])

    def get_norm(self, normalization, values, cmap):
        """Given a normalization type and values, return a normalization object.

        Parameters
        ----------
        normalization : "min_max", "0_1", "none", or "bins"
            The type of normalization to use when mapping values to colors
        values : list
            The values to use when normalizing the data
        cmap : matplotlib colormap
            The colormap to use when normalizing the data

        Returns
        -------
        matplotlib.colors normalization object
            Used to normalize data before mapping to colors
        """
        if normalization == "min_max":
            return mpc.Normalize(values[0], values[1])
        elif normalization == "0_1":
            return mpc.Normalize()
        elif normalization == "none":
            return mpc.NoNorm()
        elif normalization == "bins":
            if len(values) == (cmap.N - 1):
                return mpc.BoundaryNorm(values, cmap.N, extend="both")
            if len(values) == (cmap.N + 1):
                return mpc.BoundaryNorm(values, cmap.N)

    def get_cmap(self, cmap):
        """Converts a cmap specification to a matplotlib colormap object.

        Parameters
        ----------
            cmap : string, tuple, float, or list
                A valid mpl color, list of valid colors or a valid colormap name

        Returns
        -------
            matplotlib colormap
                a colormap matching the input
        """
        if mpc.is_color_like(cmap):
            cmap = mpc.ListedColormap([cmap])
            cmap.set_bad("grey")
            return cmap
        elif isinstance(cmap, list) and all(mpc.is_color_like(c) for c in cmap):
            cmap = mpc.ListedColormap(cmap)
            cmap.set_bad("grey")
            return cmap
        try:
            return cm.get_cmap(cmap)
        except ValueError as exception:
            print(
                "cmap must be one of: valid mpl color, list of mpl colors, or "
                f"mpl colormap:\n{str(cmap)}"
            )
            raise exception



"""Alignment objects map coordinates, vectors, and dataframes to a new sequence

Classes
-------
BaseAlignment (ABC)
    abstract base class for alignments
SequenceAlignment (BaseAlignment)
    aligns one sequence another sequence
RegionAlignment (BaseAlignment)
    cuts a sequence between a start and end position
AlignmentChain (BaseAlignment)
    allows chaining of above alignments
"""

from abc import ABC, abstractmethod
from Bio.pairwise2 import align
from Bio import SeqIO
import numpy as np
import pandas as pd
from rnavigate import data


# store sequence alignments
_alignments_cache = {}


# structure alignment parameters
# conversion of nt+pairing to pseudo amino acid sequence
def convert_sequence(aas, nts, dbn):
    """Convert pseudo-amino-acid sequence to nucleotide and dotbracket or vice versa.

    Parameters
    ----------
    aas : string or True
        the amino acid sequence
        if True, returns the amino acid translation of nts and dbn
    nts : string or True
        the nucleotide sequence
        if True, returns the nucleotide translation of aas
    dbn : string or True
        the dot-bracket notation string
        if True, returns the dot-bracket translation of aas

    Returns
    -------
    string
        sequence of the specified translation.
        If `nts` and `dbn` are True, returns a tuple.

    Example
    -------
    conver_sequence(aas="ACDEFGHIKLMNPQRSTVWY", nts=True, dbn=True)
    returns ("AAAAACCCCCUUUUUGGGGG", "([.])([.])([.])([.])")
    """
    nts_key = "AAAAACCCCCUUUUUGGGGG-"
    dbn_key = "([.])([.])([.])([.])-"
    aas_key = "ACDEFGHIKLMNPQRSTVWY-"
    to_aa = {nt + db: aa for nt, db, aa in zip(nts_key, dbn_key, aas_key)}
    to_nt = {aa: nt for nt, aa in zip(nts_key, aas_key)}
    to_db = {aa: db for db, aa in zip(dbn_key, aas_key)}
    if aas is True:
        # Make all nts uppercase and "U" instead of "T"
        nts = nts.upper().replace("T", "U")
        # Make all pseudoknots level 1 (assumes they were assigned correctly)
        pk_level = {left: "[" for left in "{<ABCDEFGHIJKLMNOPQRSTUVWXYZ"}
        pk_level |= {right: "]" for right in "}>abcdefghijklmnopqrstuvwxyz"}
        pk_level |= {keep: keep for keep in ".()[]"}
        dbn = [pk_level[db] for db in dbn]
        aas = "".join(to_aa[nt + db] for nt, db in zip(nts, dbn))
        return aas
    elif nts is True:
        nts = "".join([to_nt[aa] for aa in aas])
        return nts
    elif dbn is True:
        dbn = "".join([to_db[aa] for aa in aas])
        return dbn
    raise ValueError(
        "Please specify which sequence to return by setting aas, "
        "nts, or dbn to  True."
    )


# code used to generate structure_scoring_dict
# pair_scores = ({}
#     # both single-stranded = +6
#     | {"..": 6}
#     # pairs in the right orientation and the same nesting level = +5
#     | {pair: 5 for pair in ["((", "))", "[[", "]]"]}
#     # pairs in the right orientation and different nesting level = +2
#     | {pair: 2 for pair in ["([", "])", "[(", ")]"]}
#     # single stranded to double stranded = -8
#     | {bracket+".": -8 for bracket in "([])"}
#     | {"."+bracket: -8 for bracket in "([])"}
#     # pairs in the wrong orientation = -10
#     | {pair:-10 for pair in ["()", ")(", "[)", "(]", ")[", "](", "[]", "]["]}
#     )
# nts_key = "AAAAACCCCCUUUUUGGGGG-"
# dbn_key = "([.])([.])([.])([.])-"
# aas_key = "ACDEFGHIKLMNPQRSTVWY-"
# to_aa = {nt+db: aa for nt, db, aa in zip(nts_key, dbn_key, aas_key)}
# # matching nucleotides = +5, non-matching nucleotides = 0
# nt_scores = {n1+n2: 5  if n1==n2 else 0 for n1 in "AUCG" for n2 in "AUCG"}
# scoring_dict = {}
# for (nt1, pair1), aa1 in to_aa.items():
#     for (nt2, pair2), aa2 in to_aa.items():
#         nt_score = nt_scores[nt1+nt2]
#         pair_score = pair_scores[pair1+pair2]
#         scoring_dict[(aa1, aa2)] = nt_score + pair_score

# fmt: off
structure_scoring_dict = {
    ("A","A"):10,("A","C"):7,("A","D"):-3,("A","E"):-5,("A","F"):-5,("A","G"):5,("A","H"):2,("A","I"):-8,("A","K"):-10,("A","L"):-10,("A","M"):5,("A","N"):2,("A","P"):-8,("A","Q"):-10,("A","R"):-10,("A","S"):5,("A","T"):2,("A","V"):-8,("A","W"):-10,("A","Y"):-10,
    ("C", "A"): 7,("C", "C"): 10,("C", "D"): -3,("C", "E"): -5,("C", "F"): -5,("C", "G"): 2,("C", "H"): 5,("C", "I"): -8,("C", "K"): -10,("C", "L"): -10,("C", "M"): 2,("C", "N"): 5,("C", "P"): -8,("C", "Q"): -10,("C", "R"): -10,("C", "S"): 2,("C", "T"): 5,("C", "V"): -8,("C", "W"): -10,("C", "Y"): -10,  # noqa: E501 pylint: disable=C0301
    ("D", "A"): -3,("D", "C"): -3,("D", "D"): 11,("D", "E"): -3,("D", "F"): -3,("D", "G"): -8,("D", "H"): -8,("D", "I"): 6,("D", "K"): -8,("D", "L"): -8,("D", "M"): -8,("D", "N"): -8,("D", "P"): 6,("D", "Q"): -8,("D", "R"): -8,("D", "S"): -8,("D", "T"): -8,("D", "V"): 6,("D", "W"): -8,("D", "Y"): -8,  # noqa: E501 pylint: disable=C0301
    ("E", "A"): -5,("E", "C"): -5,("E", "D"): -3,("E", "E"): 10,("E", "F"): 7,("E", "G"): -10,("E", "H"): -10,("E", "I"): -8,("E", "K"): 5,("E", "L"): 2,("E", "M"): -10,("E", "N"): -10,("E", "P"): -8,("E", "Q"): 5,("E", "R"): 2,("E", "S"): -10,("E", "T"): -10,("E", "V"): -8,("E", "W"): 5,("E", "Y"): 2,  # noqa: E501 pylint: disable=C0301
    ("F", "A"): -5,("F", "C"): -5,("F", "D"): -3,("F", "E"): 7,("F", "F"): 10,("F", "G"): -10,("F", "H"): -10,("F", "I"): -8,("F", "K"): 2,("F", "L"): 5,("F", "M"): -10,("F", "N"): -10,("F", "P"): -8,("F", "Q"): 2,("F", "R"): 5,("F", "S"): -10,("F", "T"): -10,("F", "V"): -8,("F", "W"): 2,("F", "Y"): 5,  # noqa: E501 pylint: disable=C0301
    ("G", "A"): 5,("G", "C"): 2,("G", "D"): -8,("G", "E"): -10,("G", "F"): -10,("G", "G"): 10,("G", "H"): 7,("G", "I"): -3,("G", "K"): -5,("G", "L"): -5,("G", "M"): 5,("G", "N"): 2,("G", "P"): -8,("G", "Q"): -10,("G", "R"): -10,("G", "S"): 5,("G", "T"): 2,("G", "V"): -8,("G", "W"): -10,("G", "Y"): -10,  # noqa: E501 pylint: disable=C0301
    ("H", "A"): 2,("H", "C"): 5,("H", "D"): -8,("H", "E"): -10,("H", "F"): -10,("H", "G"): 7,("H", "H"): 10,("H", "I"): -3,("H", "K"): -5,("H", "L"): -5,("H", "M"): 2,("H", "N"): 5,("H", "P"): -8,("H", "Q"): -10,("H", "R"): -10,("H", "S"): 2,("H", "T"): 5,("H", "V"): -8,("H", "W"): -10,("H", "Y"): -10,  # noqa: E501 pylint: disable=C0301
    ("I", "A"): -8,("I", "C"): -8,("I", "D"): 6,("I", "E"): -8,("I", "F"): -8,("I", "G"): -3,("I", "H"): -3,("I", "I"): 11,("I", "K"): -3,("I", "L"): -3,("I", "M"): -8,("I", "N"): -8,("I", "P"): 6,("I", "Q"): -8,("I", "R"): -8,("I", "S"): -8,("I", "T"): -8,("I", "V"): 6,("I", "W"): -8,("I", "Y"): -8,  # noqa: E501 pylint: disable=C0301
    ("K", "A"): -10,("K", "C"): -10,("K", "D"): -8,("K", "E"): 5,("K", "F"): 2,("K", "G"): -5,("K", "H"): -5,("K", "I"): -3,("K", "K"): 10,("K", "L"): 7,("K", "M"): -10,("K", "N"): -10,("K", "P"): -8,("K", "Q"): 5,("K", "R"): 2,("K", "S"): -10,("K", "T"): -10,("K", "V"): -8,("K", "W"): 5,("K", "Y"): 2,  # noqa: E501 pylint: disable=C0301
    ("L", "A"): -10,("L", "C"): -10,("L", "D"): -8,("L", "E"): 2,("L", "F"): 5,("L", "G"): -5,("L", "H"): -5,("L", "I"): -3,("L", "K"): 7,("L", "L"): 10,("L", "M"): -10,("L", "N"): -10,("L", "P"): -8,("L", "Q"): 2,("L", "R"): 5,("L", "S"): -10,("L", "T"): -10,("L", "V"): -8,("L", "W"): 2,("L", "Y"): 5,  # noqa: E501 pylint: disable=C0301
    ("M", "A"): 5,("M", "C"): 2,("M", "D"): -8,("M", "E"): -10,("M", "F"): -10,("M", "G"): 5,("M", "H"): 2,("M", "I"): -8,("M", "K"): -10,("M", "L"): -10,("M", "M"): 10,("M", "N"): 7,("M", "P"): -3,("M", "Q"): -5,("M", "R"): -5,("M", "S"): 5,("M", "T"): 2,("M", "V"): -8,("M", "W"): -10,("M", "Y"): -10,  # noqa: E501 pylint: disable=C0301
    ("N", "A"): 2,("N", "C"): 5,("N", "D"): -8,("N", "E"): -10,("N", "F"): -10,("N", "G"): 2,("N", "H"): 5,("N", "I"): -8,("N", "K"): -10,("N", "L"): -10,("N", "M"): 7,("N", "N"): 10,("N", "P"): -3,("N", "Q"): -5,("N", "R"): -5,("N", "S"): 2,("N", "T"): 5,("N", "V"): -8,("N", "W"): -10,("N", "Y"): -10,  # noqa: E501 pylint: disable=C0301
    ("P", "A"): -8,("P", "C"): -8,("P", "D"): 6,("P", "E"): -8,("P", "F"): -8,("P", "G"): -8,("P", "H"): -8,("P", "I"): 6,("P", "K"): -8,("P", "L"): -8,("P", "M"): -3,("P", "N"): -3,("P", "P"): 11,("P", "Q"): -3,("P", "R"): -3,("P", "S"): -8,("P", "T"): -8,("P", "V"): 6,("P", "W"): -8,("P", "Y"): -8,  # noqa: E501 pylint: disable=C0301
    ("Q", "A"): -10,("Q", "C"): -10,("Q", "D"): -8,("Q", "E"): 5,("Q", "F"): 2,("Q", "G"): -10,("Q", "H"): -10,("Q", "I"): -8,("Q", "K"): 5,("Q", "L"): 2,("Q", "M"): -5,("Q", "N"): -5,("Q", "P"): -3,("Q", "Q"): 10,("Q", "R"): 7,("Q", "S"): -10,("Q", "T"): -10,("Q", "V"): -8,("Q", "W"): 5,("Q", "Y"): 2,  # noqa: E501 pylint: disable=C0301
    ("R", "A"): -10,("R", "C"): -10,("R", "D"): -8,("R", "E"): 2,("R", "F"): 5,("R", "G"): -10,("R", "H"): -10,("R", "I"): -8,("R", "K"): 2,("R", "L"): 5,("R", "M"): -5,("R", "N"): -5,("R", "P"): -3,("R", "Q"): 7,("R", "R"): 10,("R", "S"): -10,("R", "T"): -10,("R", "V"): -8,("R", "W"): 2,("R", "Y"): 5,  # noqa: E501 pylint: disable=C0301
    ("S", "A"): 5,("S", "C"): 2,("S", "D"): -8,("S", "E"): -10,("S", "F"): -10,("S", "G"): 5,("S", "H"): 2,("S", "I"): -8,("S", "K"): -10,("S", "L"): -10,("S", "M"): 5,("S", "N"): 2,("S", "P"): -8,("S", "Q"): -10,("S", "R"): -10,("S", "S"): 10,("S", "T"): 7,("S", "V"): -3,("S", "W"): -5,("S", "Y"): -5,  # noqa: E501 pylint: disable=C0301
    ("T", "A"): 2,("T", "C"): 5,("T", "D"): -8,("T", "E"): -10,("T", "F"): -10,("T", "G"): 2,("T", "H"): 5,("T", "I"): -8,("T", "K"): -10,("T", "L"): -10,("T", "M"): 2,("T", "N"): 5,("T", "P"): -8,("T", "Q"): -10,("T", "R"): -10,("T", "S"): 7,("T", "T"): 10,("T", "V"): -3,("T", "W"): -5,("T", "Y"): -5,  # noqa: E501 pylint: disable=C0301
    ("V", "A"): -8,("V", "C"): -8,("V", "D"): 6,("V", "E"): -8,("V", "F"): -8,("V", "G"): -8,("V", "H"): -8,("V", "I"): 6,("V", "K"): -8,("V", "L"): -8,("V", "M"): -8,("V", "N"): -8,("V", "P"): 6,("V", "Q"): -8,("V", "R"): -8,("V", "S"): -3,("V", "T"): -3,("V", "V"): 11,("V", "W"): -3,("V", "Y"): -3,  # noqa: E501 pylint: disable=C0301
    ("W", "A"): -10,("W", "C"): -10,("W", "D"): -8,("W", "E"): 5,("W", "F"): 2,("W", "G"): -10,("W", "H"): -10,("W", "I"): -8,("W", "K"): 5,("W", "L"): 2,("W", "M"): -10,("W", "N"): -10,("W", "P"): -8,("W", "Q"): 5,("W", "R"): 2,("W", "S"): -5,("W", "T"): -5,("W", "V"): -3,("W", "W"): 10,("W", "Y"): 7,  # noqa: E501 pylint: disable=C0301
    ("Y", "A"): -10,("Y", "C"): -10,("Y", "D"): -8,("Y", "E"): 2,("Y", "F"): 5,("Y", "G"): -10,("Y", "H"): -10,("Y", "I"): -8,("Y", "K"): 2,("Y", "L"): 5,("Y", "M"): -10,("Y", "N"): -10,("Y", "P"): -8,("Y", "Q"): 2,("Y", "R"): 5,("Y", "S"): -5,("Y", "T"): -5,("Y", "V"): -3,("Y", "W"): 7,("Y", "Y"): 10,  # noqa: E501 pylint: disable=C0301
}
# fmt: off


def set_alignment(
    sequence1,
    sequence2,
    alignment1,
    alignment2,
    t_or_u="U",
):
    """Add an alignment to be used as the default between two sequences.

    When objects with these sequences are aligned for visualization, RNAvigate
    uses this alignment instead of an automated pairwise sequence alignment.
    Alignment 1 and 2 must have matching lengths.
    alignment(1,2) and sequence(1,2) must differ only by dashes "-".

    e.g.:
        sequence1 ="AAGCUUCGGUACAUGCAAGAUGUAC"
        sequence2 ="AUCGAUCGAGCUGCUGUGUACGUAC"
        alignment1="AAGCUUCG---------GUACAUGCAAGAUGUAC"
        alignment2="AUCGAUCGAGCUGCUGUGUAC---------GUAC"
                     |mm|   | indel |    | indel |

    Parameters
    ----------
    sequence1 : string
        the first sequence
    sequence2 : string
        the second sequence
    alignment1 : string
        first sequence, plus dashes "-" indicating indels
    alignment2 : string
        second sequence, plus dashes "-" indicating indels
    t_or_u : "T", "U", or False
        "T" converts "U"s to "T"s
    """
    # Normalize sequences
    sequence1 = data.normalize_sequence(sequence1, t_or_u=t_or_u)
    sequence2 = data.normalize_sequence(sequence2, t_or_u=t_or_u)
    hash1 = hash(sequence1)
    hash2 = hash(sequence2)
    alignment1 = data.normalize_sequence(alignment1, t_or_u=t_or_u)
    alignment2 = data.normalize_sequence(alignment2, t_or_u=t_or_u)

    # Check for consistency
    if len(alignment1) != len(alignment2):
        raise ValueError(
            "rnavigate.data.set_alignment:\n"
            "    Alignment 1 and 2 must having matching length.\n"
            f"    alignment 1: {len(alignment1):>7}\n"
            f"    alignment 2: {len(alignment2):>7}\n"
        )
    if alignment1.replace("-", "") != sequence1:
        raise ValueError("Alignment 1 does not match sequence 1")
    if alignment2.replace("-", "") != sequence2:
        raise ValueError("Alignment 2 does not match sequence 2")
    # Store dictionaries in _alignments_cache
    if hash1 not in _alignments_cache and hash2 not in _alignments_cache:
        _alignments_cache[hash1] = {hash2: {"seqA": alignment1, "seqB": alignment2}}
    elif hash1 in _alignments_cache:
        _alignments_cache[hash1].update(
            {hash2: {"seqA": alignment1, "seqB": alignment2}}
        )
    elif hash2 in _alignments_cache:
        _alignments_cache[hash2].update(
            {hash1: {"seqA": alignment2, "seqB": alignment1}}
        )


def set_multiple_sequence_alignment(fasta, set_pairwise=False):
    """Set alignments from a multiple sequence alignment Pearson fasta file.

    Sets alignments to a base sequence, then returns the base sequence to be
    when a multiple sequence alignment plot is desired. Also sets all pairwise
    alignments, if desired. When setting pairwise alignments, dashes that are
    shared between pairwise sequences are removed first.

    Parameters
    ----------
    fasta : string
        location of Pearson fasta file
    set_pairwise : bool, defaults to False
        whether to set every pairwise alignment as well as the multiple
        sequence alignment.
    """
    with open(fasta, "r") as file:
        fasta = list(SeqIO.parse(file, "fasta"))
        fasta = [str(seq.seq).upper().replace("T", "U") for seq in fasta]
    base_sequence = []
    for nts in zip(*fasta):
        nts = [nt for nt in nts if nt != "-"]
        most_frequent = max(set(nts), key=nts.count)
        base_sequence.append(most_frequent)
    base_sequence = "".join(base_sequence)
    for alignment in fasta:
        sequence = alignment.replace("-", "")
        set_alignment(base_sequence, sequence, base_sequence, alignment)
    if not set_pairwise:
        return data.Sequence(base_sequence)
    for i, seq1 in enumerate(fasta[:-1]):
        for seq2 in fasta[i + 1 :]:
            alignment1, alignment2 = [], []
            for nt1, nt2 in zip(seq1, seq2):
                if nt1 == "-" and nt2 == "-":
                    continue
                alignment1.append(nt1)
                alignment2.append(nt2)
            alignment1 = "".join(alignment1)
            alignment2 = "".join(alignment2)
            sequence1 = alignment1.replace("-", "")
            sequence2 = alignment2.replace("-", "")
            set_alignment(sequence1, sequence2, alignment1, alignment2)
    return data.Sequence(base_sequence)


def lookup_alignment(sequence1, sequence2, t_or_u="U"):
    """look up a previously set alignment in the _alignments_cache

    Parameters
    ----------
    sequence1 : string
        The first sequence to align
    sequence2 : string
        The second sequence to be aligned to
    t_or_u : "T", "U", or False, defaults to "U"
        "T" converts "U"s to "T"s
        "U" converts "U"s to "T"s
        False does nothing

    Returns
    -------
    dictionary, if an alignment is found, otherwise None
        {"seqA": sequence1 with gap characters representing alignment,
         "seqB": sequence2 with gap characters representing alignment}
    """
    # Normalize sequences
    sequence1 = data.normalize_sequence(sequence1, t_or_u=t_or_u)
    sequence2 = data.normalize_sequence(sequence2, t_or_u=t_or_u)
    # create hash keys
    hash1 = hash(sequence1)
    hash2 = hash(sequence2)
    # lookup and return alignment for sequence1 to sequence2
    try:
        return _alignments_cache[hash1][hash2]
    except KeyError:
        try:
            inverse_alignment = _alignments_cache[hash2][hash1]
            return {
                "seqA": inverse_alignment["seqB"],
                "seqB": inverse_alignment["seqA"],
            }
        except KeyError:
            return None


class BaseAlignment(ABC):
    """Abstract base class for alignments

    Parameters
    ----------
    starting_sequence : string
        the sequence to be aligned
    target_length : int
        the length of the target sequence

    Attributes
    ----------
    starting_sequence : string
        the beginning sequence
    mapping : numpy.array
        the alignment map array.
        index of starting_sequence is mapping[index] of target_sequence
    target_sequence : string
        the portion of starting sequence that is mapped
    target_length : integer
        the length of the target sequence
    """

    def __init__(self, starting_sequence, target_length):
        """Creates a BaseAlignment with starting and target sequences."""
        self.starting_sequence = starting_sequence
        self.mapping = self.get_mapping()
        self.target_length = target_length
        self.target_sequence = self.get_target_sequence()

    @abstractmethod
    def get_mapping(self):
        """Alignments require a mapping from starting to target sequence"""
        return

    @abstractmethod
    def get_inverse_alignment(self):
        """Alignments require a method to get the inverted alignment"""
        return

    def get_target_sequence(self):
        """Gets the portion of starting sequence that fits the alignment"""
        return "".join(self.map_values(list(self.starting_sequence), "."))

    def map_values(self, values, fill=np.nan):
        """Takes an array of length equal to starting sequence and maps them to
        target sequence, unmapped positions in starting sequence are dropped
        and unmapped positions in target sequence are filled with fill value.

        Parameters
        ----------
        values : iterable
            values to map to target sequence.
        fill : any, defaults to np.nan
            a value for unmapped positions in target sequence.

        Returns
        -------
        numpy.array
            an array of values equal in length to target sequence
        """
        new_values = np.full(self.target_length, fill)
        for idx1, value in enumerate(values):
            idx2 = self.mapping[idx1]
            if idx2 == -1:
                continue
            new_values[idx2] = value
        return new_values

    def map_indices(self, indices, keep_minus_one=True):
        """Takes a list of indices (0-index) and maps them to target sequence

        Parameters
        ----------
        indices : int or list of int
            a single or list of integer indices
        keep_minus_one : bool, defaults to True
            whether to keep unmapped starting sequence indices (-1) in the
            returned array.

        Returns
        -------
        numpy.array
            the equivalent indices in target sequence
        """
        indices = np.array(indices, dtype=int)
        new_indices = self.mapping[indices]
        if not keep_minus_one:
            new_indices = new_indices[new_indices != -1]
        return new_indices

    def map_positions(self, positions, keep_zero=True):
        """Takes a list of positions (1-index) and maps them to target sequence

        Parameters
        ----------
        positions : int or list of int
            a single or list of integer positions
        keep_zero : bool, defaults to True
            whether to keep unmapped starting sequence positions (0) in the
            returned array.

        Returns
        -------
        numpy.array
            the equivalent positions in target sequence
        """
        positions = np.array(positions, dtype=int)
        new_positions = self.mapping[positions - 1] + 1
        if not keep_zero:
            new_positions = new_positions[new_positions != 0]
        return new_positions

    def map_dataframe(self, dataframe, position_columns):
        """Takes a dataframe and maps position columns to target sequence.

        Rows with unmapped positions are dropped.

        Parameters
        ----------
        dataframe : pandas.DataFrame
            a dataframe with position columns
        position_columns : list of str
            a list of columns containing positions to map

        Returns
        -------
        pandas.DataFrame
            a new dataframe (copy) with position columns mapped or dropped
        """
        dataframe = dataframe.copy()
        for col in position_columns:
            dataframe[col] = self.map_positions(dataframe[col].values)
            dataframe = dataframe[dataframe[col] != 0]
        return dataframe.copy()

    def map_nucleotide_dataframe(
        self, dataframe, position_column="Nucleotide", sequence_column="Sequence"
    ):
        """Takes a per-nt dataframe and map it to the target sequence.

        Dataframe must have 1 row per nucleotide in starting sequence,
        with a position column and a sequence column. Dataframe is
        mapped to have the same format, but for target sequence nucleotides and
        positions.

        Parameters
        ----------
        dataframe : pandas.DataFrame
            a per-nucleotide dataframe
        position_column : string, defaults to "Nucleotide"
            name of the position column.
        sequence_column : string, defaults to "Sequence"
            name of the sequence column.

        Returns
        -------
        pandas.DataFrame
            a new dataframe (copy) mapped to target sequence.
            Unmapped starting sequence positions are dropped and unmapped
            target sequence positions are filled.
        """
        dataframe = dataframe.copy()
        dataframe[position_column] = self.map_positions(dataframe[position_column])
        dataframe = dataframe[dataframe[position_column] != 0]
        new_dataframe = pd.DataFrame(
            {
                position_column: np.arange(self.target_length) + 1,
            }
        )
        new_dataframe = new_dataframe.merge(dataframe, "left", position_column)
        new_dataframe[sequence_column] = list(self.target_sequence)
        return new_dataframe


class SequenceAlignment(BaseAlignment):
    """The most useful feature of RNAvigate. Maps positions from one sequence
    to a totally different sequence using user-defined pairwise alignment or
    automatic pairwise alignment.

    Parameters
    ----------
    sequence1 : string
        the sequence to be aligned
    sequence2 : string
        the sequence to align to
    align_kwargs : dict, defaults to None
        a dictionary of arguments to pass to pairwise2.align.globalms
    full : bool, defaults to False
        whether to keep unmapped starting sequence positions.
    use_previous : bool, defaults to True
        whether to use previously set alignments

    Attributes
    ----------
    sequence1 : str
        the sequence to be aligned
    sequence2 : str
        the sequence to align to
    alignment1 : str
        the alignment string matching sequence1 to sequence2
    alignment2 : str
        the alignment string matching sequence2 to sequence1
    starting_sequence : str
        sequence1
    target_sequence : str
        sequence2 if full is False, else alignment2
    mapping : numpy.array
        the alignment map array.
        index of starting_sequence is mapping[index] of target_sequence
    """

    def __init__(
        self, sequence1, sequence2, align_kwargs=None, full=False, use_previous=True
    ):
        """Creates an alignment from sequence1 to sequence2."""
        if align_kwargs is None:
            self.align_kwargs = {"match": 1, "mismatch": 0, "open": -5, "extend": -0.1}
        if isinstance(sequence1, data.Sequence):
            sequence1 = sequence1.sequence
        if isinstance(sequence2, data.Sequence):
            sequence2 = sequence2.sequence
        self.sequence1 = sequence1
        self.sequence2 = sequence2
        self.alignment1, self.alignment2 = self.get_alignment()
        self.full = full
        self.use_previous = use_previous
        if self.full:
            target_length = len(self.alignment2)
        else:
            target_length = len(self.sequence2)
        super().__init__(sequence1, target_length)

    def __repr__(self):
        """a nice text only representation of an alignment"""
        al1, al2 = self.alignment1, self.alignment2
        misses = "".join([" ", "X"][i != j] for i, j in zip(al1, al2))
        position = "".join(f"{n:<20}" for n in range(1, len(al1), 20))
        return (
            "alignment:\n"
            f"    {position}\n"
            f"    {al1}\n"
            f"    {misses}\n"
            f"    {al2}\n"
        )

    def print(self, print_format="full"):
        """Print the alignment in a human-readable format.

        Parameters
        ----------
        print_format : "full", "cigar", "long" or "short", defaults to "full"
            how to format the alignment.
            "full": the full length alignment with changes labeled "X"
            "cigar": the CIGAR string
            "long": locations and sequences of each change
            "short": total number of matches, mismatches, and indels
        """
        if print_format == "full":
            print(self)
        elif print_format == "cigar":
            self.print_cigar()
        elif print_format == "short":
            self.print_number_of_changes()
        elif print_format == "long":
            self.print_all_changes()

    def print_cigar(self):
        """Print the CIGAR string"""
        alignment1 = np.array(list(self.alignment1))
        alignment2 = np.array(list(self.alignment2))
        cigar_string = ""
        current_tag = ""
        current_count = 0
        for nt1, nt2 in zip(alignment1, alignment2):
            if nt1 == nt2:
                this_tag = "="
            elif nt1 == "-":
                this_tag = "I"
            elif nt2 == "-":
                this_tag = "D"
            elif nt1 != nt2:
                this_tag = "X"
            if current_tag == "":
                current_tag = this_tag
                current_count += 1
            elif current_tag == this_tag:
                current_count += 1
            elif current_tag != this_tag:
                cigar_string += str(current_count) + current_tag
                current_tag = this_tag
                current_count = 1
        cigar_string += str(current_count) + current_tag
        print(f"    {cigar_string}\n")

    def print_number_of_changes(self):
        """Print the total numbers of matches, mismatches, and indels."""
        alignment1 = np.array(list(self.alignment1))
        alignment2 = np.array(list(self.alignment2))
        tags = {}
        tags["deletions"] = np.sum(alignment2 == "-")
        tags["insertions"] = np.sum(alignment1 == "-")
        tags["matches"] = np.sum(alignment1 == alignment2)
        tags["mismatches"] = (
            np.sum(alignment1 != alignment2) - tags["deletions"] - tags["insertions"]
        )
        for key in ["matches", "mismatches", "deletions", "insertions"]:
            print(f"    {key:<10} {tags[key]}")
        print()

    def print_all_changes(self):
        """Print location and sequence of all changes."""

        def print_line(tag, start, seq1, seq2):
            if tag == "match":
                return
            elif tag == "mismatch":
                string = f"    {tag:<9} {start:>6} {seq1} --> {seq2}"
            elif tag == "delete":
                string = f"    {tag:<9} {start:>6} {seq1}"
            elif tag == "insert":
                string = f"    {tag:<9} {start:>6} {seq2}"
            print(string)

        alignment1 = np.array(list(self.alignment1))
        alignment2 = np.array(list(self.alignment2))
        current_tag = ""
        current_start = 0
        this_pos = 0
        seq1, seq2 = "", ""
        for nt1, nt2 in zip(alignment1, alignment2):
            if nt1 != "-":
                this_pos += 1
            if nt1 == nt2:
                this_tag = "match"
            elif nt1 == "-":
                this_tag = "insert"
            elif nt2 == "-":
                this_tag = "delete"
            elif nt1 != nt2:
                this_tag = "mismatch"
            if current_tag == "":
                current_tag = this_tag
                current_start = this_pos
                seq1, seq2 = nt1, nt2
            elif current_tag == this_tag:
                seq1 += nt1
                seq2 += nt2
            elif current_tag != this_tag:
                print_line(current_tag, current_start, seq1, seq2)
                current_tag = this_tag
                current_start = this_pos
                seq1, seq2 = nt1, nt2
        print_line(current_tag, current_start, seq1, seq2)
        print()

    def get_inverse_alignment(self):
        """Gets an alignment that maps from sequence2 to sequence1."""
        return SequenceAlignment(self.sequence2, self.sequence1, self.full)

    def get_alignment(self):
        """Gets an alignment that has either been user-defined or previously
        calculated or produces a new pairwise alignment between two sequences.

        Returns
        -------
        alignment1, alignment2 : tuple of 2 str
            the alignment strings matching sequence1 and sequence2, respectively.
        """
        # Normalize sequences
        seq1 = self.sequence1.upper().replace("T", "U")
        seq2 = self.sequence2.upper().replace("T", "U")
        # Check if sequences match
        if seq1 == seq2:
            return (seq1, seq2)
        # check for sequences that differ only at dashes
        a1 = np.array(list(seq1))
        a2 = np.array(list(seq2))
        if len(a1) == len(a2):
            matches = sum((a1 == a2) | (a1 == "-") | (a2 == "-"))
            if matches == len(a1):
                return (seq1, seq2)
        # if not already set, do a pairwise alignment and set
        alignments = lookup_alignment(seq1, seq2)
        if alignments is None:
            alignment = align.globalms(
                seq1,
                seq2,
                penalize_end_gaps=False,
                one_alignment_only=True,
                **self.align_kwargs,
            )
            set_alignment(
                sequence1=seq1,
                sequence2=seq2,
                alignment1=alignment[0].seqA,
                alignment2=alignment[0].seqB,
            )
            alignments = lookup_alignment(seq1, seq2)
        align1, align2 = alignments["seqA"], alignments["seqB"]
        return (align1, align2)

    def get_mapping(self):
        """Calculates a mapping from starting sequence to target sequence.

        Returns
        -------
        mapping : numpy.array
            an array that maps to an index of target sequence.
            index of starting_sequence is mapping[index] of target_sequence
        """
        align1 = self.alignment1
        align2 = self.alignment2
        # get an index mapping from sequence 1 to the full alignment
        seq1_to_align = np.where([nt != "-" for nt in align1])[0]
        # if we want a mapping to a position in the full alignment, this is it.
        if self.full:
            return seq1_to_align
        # extra steps to get to sequence 2 positions
        # positions that are removed when plotting on sequence 2
        align_mask = np.array([nt != "-" for nt in align2])
        # an index mapping from the full alignment to position in sequence 2
        align_to_seq2 = np.full(len(align2), -1)
        align_to_seq2[align_mask] = np.arange(len(self.sequence2))
        # an index mapping from sequence 1 to sequence 2 positions
        seq1_to_seq2 = align_to_seq2[seq1_to_align]
        return seq1_to_seq2


class AlignmentChain(BaseAlignment):
    """Combines a list of alignments into one.

    Parameters
    ----------
    alignments : list of Alignment objects
        the alignments to chain together

    Attributes
    ----------
    alignments : list
        the constituent alignments
    starting_sequence : str
        starting sequence of alignments[0]
    target_sequence : str
        target sequence of alignments[-1]
    mapping : numpy.array
        an array which maps from `starting_sequence` to `target_sequence`.
        index of starting_sequence is mapping[index] of target sequence
    """

    def __init__(self, *alignments):
        """Creates a single alignment from multiple alignments."""
        next_sequence_len = len(alignments[0].starting_sequence)
        for alignment in alignments:
            if next_sequence_len == len(alignment.starting_sequence):
                next_sequence_len = len(alignment.target_sequence)
            else:
                raise ValueError("Alignments do not chain.")

        self.alignments = alignments
        super().__init__(
            self.alignments[0].starting_sequence,
            len(self.alignments[-1].target_sequence),
        )

    def get_mapping(self):
        """combines mappings from each alignment.

        Returns
        -------
        mapping : numpy.array
            mapping from initial starting sequence to final target sequence
            index of starting_sequence is mapping[index] of target sequence
        """
        indices = self.alignments[0].mapping
        for alignment in self.alignments[1:]:
            valid = indices != -1
            indices[valid] = alignment.map_indices(indices[valid])
        return indices

    def get_inverse_alignment(self):
        alignments = [a.get_inverse_alignment() for a in self.alignments[::-1]]
        return AlignmentChain(*alignments)


class StructureAlignment(BaseAlignment):
    """Experimental secondary structure alignment based on RNAlign2D algorithm
    (https://doi.org/10.1186/s12859-021-04426-8)

    Parameters
    ----------
    sequence1 : string
        the sequence to be aligned
    sequence2 : string
        the sequence to align to
    structure1 : string, defaults to None
        the secondary structure of sequence1
    structure2 : string, defaults to None
        the secondary structure of sequence2
    full : bool, defaults to False
        whether to align to full length of sequence2 or just mapped length

    Attributes
    ----------
    sequence1 : str
        the sequence to be aligned
    sequence2 : str
        the sequence to align to
    structure1 : str
        the secondary structure of sequence1
    structure2 : str
        the secondary structure of sequence2
    alignment1 : str
        the alignment string matching sequence1 to sequence2
    alignment2 : str
        the alignment string matching sequence2 to sequence1
    starting_sequence : str
        sequence1
    target_sequence : str
        sequence2 if full is False, else alignment2
    mapping : numpy.array
        the alignment map array.
        index of starting_sequence is mapping[index] of target_sequence
    """

    def __init__(
        self, sequence1, sequence2, structure1=None, structure2=None, full=False
    ):
        """Creates an alignment from structure1 to structure2."""
        if structure1 is None:
            structure1 = sequence1
        if structure2 is None:
            structure2 = sequence2
        if isinstance(sequence1, data.Sequence):
            sequence1 = sequence1.sequence
        if isinstance(structure1, data.SecondaryStructure):
            structure1 = structure1.get_dotbracket()
        self.sequence1 = sequence1
        self.structure1 = structure1
        if isinstance(sequence2, data.Sequence):
            sequence2 = sequence2.sequence
        if isinstance(structure2, data.SecondaryStructure):
            structure2 = structure2.get_dotbracket()
        self.sequence2 = sequence2
        self.structure2 = structure2
        self.alignment1, self.alignment2 = self.get_alignment()
        self.full = full
        if self.full:
            target_length = len(self.alignment2)
        else:
            target_length = len(self.sequence2)
        super().__init__(self.sequence1, target_length)

    def get_alignment(self):
        """Aligns pseudo-amino-acid sequences according to RNAlign2D rules.

        Returns
        -------
        alignment1, alignment2 : tuple of 2 str
            the alignment strings matching sequence1 and sequence2, respectively.
        """
        # Normalize sequences
        seq1 = convert_sequence(aas=True, nts=self.sequence1, dbn=self.structure1)
        seq2 = convert_sequence(aas=True, nts=self.sequence2, dbn=self.structure2)
        # Check if sequences match
        if seq1 == seq2:
            return (seq1, seq2)
        else:
            alignment = align.globalds(
                seq1,
                seq2,
                structure_scoring_dict,
                -12,
                -1,
                penalize_end_gaps=False,
                one_alignment_only=True,
            )[0]
            alignment1 = alignment.seqA
            alignment2 = alignment.seqB
            set_alignment(
                sequence1=seq1,
                sequence2=seq2,
                alignment1=alignment1,
                alignment2=alignment2,
                t_or_u=False,
            )
            alignment = lookup_alignment(seq1, seq2, t_or_u=False)
            alignment1, alignment2 = alignment["seqA"], alignment["seqB"]
        # convert pseudo-amino acid alignments back into nucleotide alignments
        return alignment1, alignment2

    def get_mapping(self):
        """Calculates a mapping from starting sequence to target sequence.

        Returns
        -------
        mapping : numpy.array
            an array which maps an indices to the target sequence.
            starting_sequence[idx] == target_sequence[self.mapping[idx]]
        """
        align1 = self.alignment1
        align2 = self.alignment2
        # get an index mapping from sequence 1 to the full alignment
        seq1_to_align = np.where([nt != "-" for nt in align1])[0]
        # if we want a mapping to a position in the full alignment, this is it.
        if self.full:
            return seq1_to_align
        # extra steps to get to sequence 2 positions
        # positions that are removed when plotting on sequence 2
        align_mask = np.array([nt != "-" for nt in align2])
        # an index mapping from the full alignment to position in sequence 2
        align_to_seq2 = np.full(len(align2), -1)
        align_to_seq2[align_mask] = np.arange(len(self.sequence2))
        # an index mapping from sequence 1 to sequence 2 positions
        seq1_to_seq2 = align_to_seq2[seq1_to_align]
        return seq1_to_seq2

    def get_inverse_alignment(self):
        """Gets an alignment that maps from sequence2 to sequence1."""
        return StructureAlignment(
            self.sequence2, self.sequence1, self.structure2, self.structure1
        )

    def set_as_default_alignment(self):
        """Set this as the default alignment between sequence1 and sequence2."""
        set_alignment(
            sequence1=self.sequence1,
            sequence2=self.sequence2,
            alignment1=convert_sequence(aas=self.alignment1, nts=True, dbn=False),
            alignment2=convert_sequence(aas=self.alignment2, nts=True, dbn=False),
        )



#!/usr/bin/env python

###############################################################################
#  SecondaryStructure code originally based on RNAtools.py
#      Main contributors:
#           Gregg Rice
#           Anthony Mustoe
#           Tom Christy
#           Patrick Irving
###############################################################################

import sys
from os.path import isfile
import numpy as np
import pandas as pd
import math
import xml.etree.ElementTree as xmlet
import json
from rnavigate import data


class SecondaryStructure(data.Sequence):
    """Base class for secondary structures.

    Parent classes:
        rnav.data.Data(rnav.data.Sequence)

    Attributes:
        data (str | pandas.DataFrame): dataframe storing base-pairs
            One row for every nucleotide position
            Required columns:
                "Nucleotide"   the nucleotide position
                "Sequence"     the nucleotide letter
                "Pair"         the nucleotide position of the base-pair
                               0 indicates a single stranded nucleotide
            Optional columns:
                "X_coordinate" the x position of the nucleotide in the drawing
                "Y_coordinate" the y position of the nucleotide in the drawing
        sequence (str): sequence string
        nts (numpy.array): "Nucleotide" column of data
        pair_nts (numpy.array): "Pair" column of data
        header (str): header information from CT file
        xcoordinates (numpy array): "X_coordinate" column of data
        ycoordinates (numpy array): "X_coordinate" column of data
    """

    ###########################################################################
    # Initialization methods
    ###########################################################################

    def __init__(self, input_data, extension=None, autoscale=True, name=None, **kwargs):
        """Creates a SecondaryStructure object from a given file or dataframe.

        Args:
            input_data (str | pandas.DataFrame):
                A dataframe or filepath containing a secondary structure
                DataFrame should contain these columns:
                    ["Nucleotide", "Sequence", "Pair"]
                "Pair" column must be redundant.
                Filepath parsing is determined by file extension:
                    varna, xrna, nsd, cte, ct, dbn, bracket, json (R2DT), forna
        """
        if isinstance(input_data, pd.DataFrame):
            self.data = input_data
            self.filepath = "dataframe"
        elif isfile(input_data):
            if extension is None:
                extension = input_data.split(".")[-1].lower()
            read_file = {
                "varna": self.read_varna,
                "xrna": self.read_xrna,
                "nsd": self.read_nsd,
                "cte": self.read_cte,
                "ct": self.read_ct,
                "db": self.read_dotbracket,
                "dbn": self.read_dotbracket,
                "bracket": self.read_dotbracket,
                "json": self.read_r2dt,
                "forna": self.read_forna,
            }[extension]
            self.filepath = input_data
            self.data = read_file(**kwargs)
        super().__init__(input_data=self.data, name=name)
        if "X_coordinate" in self.data.columns and autoscale is True:
            self.transform_coordinates(scale=1, center=(0, 0))
        self.distance_matrix = None

    @classmethod
    def from_sequence(cls, input_data):
        seq = data.Sequence(input_data=input_data)
        df = pd.DataFrame(
            data={
                "Nucleotide": np.arange(seq.length) + 1,
                "Sequence": list(seq.sequence),
                "Pair": np.zeros(seq.length),
            }
        )
        df = df.astype(
            dtype={"Nucleotide": "Int32", "Sequence": "string", "Pair": "Int32"}
        )
        return cls(input_data=df)

    @classmethod
    def from_pairs_list(cls, input_data, sequence):
        """Creates a SecondaryStructure from a list of pairs and a sequence.

        Args:
            pairs (list): 1-indexed list of base pairs. e.g. [(1, 20), (2, 19)]
            sequence (str): sequence string. e.g., "AUCGUGUCAUGCUA"
        """
        structure = cls.from_sequence(sequence)
        structure.add_pairs(input_data)
        return structure

    ###########################################################################
    # properties
    ###########################################################################

    @property
    def nts(self):
        return self.data["Nucleotide"].values

    @property
    def pair_nts(self):
        return self.data["Pair"].values

    @property
    def ycoordinates(self):
        return self.data["Y_coordinate"].values

    @ycoordinates.setter
    def ycoordinates(self, values):
        self.data["Y_coordinate"] = values

    @property
    def xcoordinates(self):
        return self.data["X_coordinate"].values

    @xcoordinates.setter
    def xcoordinates(self, values):
        self.data["X_coordinate"] = values

    def __str__(self):
        """print the filepath and length of the RNA"""
        return f"Name = {self.filepath}, length = {len(self.pair_nts)}"

    ###########################################################################
    # Loading Data
    ###########################################################################

    def read_ct(self, structure_number=0):
        """Loads secondary structure information from a given ct file.

        Requires a properly formatted header!

        Args:
            structure_number (int, optional): If ct file contains multiple structures,
                uses the given structure. Defaults to 0.
        """
        fIN = self.filepath
        num, seq, bp, mask = [], "", [], []
        try:
            with open(fIN) as f:
                # process the first line (pops off the first element)
                spl = f.readline().split()
                nt_length = int(spl[0])
                header = " ".join(spl[1:])
                current_structure = 0
                for i, line in enumerate(f):
                    spl = line.split()
                    if (i + 1) % (nt_length + 1) == 0:
                        # catch the next header
                        current_structure += 1
                        if current_structure > structure_number:
                            break
                        header = " ".join(spl[1:])
                        continue
                    if current_structure == structure_number:
                        num.append(int(spl[0]))
                        seq += str(spl[1])
                        bp.append(int(spl[4]))
                        # check if there is masking info
                        if len(spl) == 7 and spl[6] == "1":
                            mask.append(1)
                        else:
                            mask.append(0)
        except Exception as e:
            raise IOError(f"{fIN} has invalid format or does not exist") from e
        if len(num) == 0:
            sys.exit(f"Structure {structure_number} was not found in the ct file")
        # check consistency!
        for idx, pair in enumerate(bp):
            if pair == 0:
                continue
            p1 = idx + 1
            p2 = bp[pair - 1]
            if p1 != p2:
                print(f"WARNING: inconsistency: {pair} paired to {p1} and {p2}")
        self.header = header
        return pd.DataFrame(
            {"Nucleotide": num, "Sequence": list(seq), "Pair": bp, "Mask": mask}
        )

    def read_varna(self):
        """Generates SecondaryStructure object data from a VARNA file,
        including nucleotide x and y coordinates.
        """
        # Parse file and get sequence, xcoords, ycoords, and list of pairs.
        tree = xmlet.parse(self.filepath)
        root = tree.getroot()
        # extract sequence, y and x coordinates
        sequence, xcoords, ycoords = "", [], []
        for nt in root.findall("./RNA/bases/nt"):
            base = nt.find("base").text.strip()
            sequence += base
            for i in nt:
                if i.get("r") == "pos":
                    xcoords.append(float(i.get("x")))
                    ycoords.append(float(i.get("y")))
        # extract pairing information
        basepairs = []
        for pair in root.findall("./RNA/BPs/bp"):
            i = int(pair.get("part5")) + 1
            j = int(pair.get("part3")) + 1
            basepairs.append((i, j))
        # store attributes
        pairs = [0] * len(sequence)
        for i, j in basepairs:
            pairs[i - 1] = j
            pairs[j - 1] = i
        return pd.DataFrame(
            {
                "Nucleotide": np.arange(len(sequence)) + 1,
                "Sequence": list(sequence),
                "Pair": pairs,
                "X_coordinate": xcoords,
                "Y_coordinate": ycoords,
            }
        )

    def read_xrna(self):
        """Generates SecondaryStructure object data from an XRNA file,
        including nucleotide x and y coordinates.
        """
        # Parse file and get sequence, xcoords, ycoords, and list of pairs.
        tree = xmlet.parse(self.filepath)
        root = tree.getroot()
        # extract sequence, x and y coordinates
        nucList = root.findall("./Complex/RNAMolecule/")
        nucLists = []
        for i in nucList:
            if i.tag == "NucListData":
                nucLists.append(i)
        sequence, xcoords, ycoords = "", [], []
        for nt in nucLists[0].text.split("\n"):
            if nt == "":
                continue
            line = nt.split()
            sequence += line[0]
            xcoords.append(float(line[1]))
            ycoords.append(float(line[2]))
        # extract pairing information
        basepairs = []
        for helix in root.findall("./Complex/RNAMolecule/BasePairs"):
            i_outter = int(helix.get("nucID"))
            j_outter = int(helix.get("bpNucID"))
            length = int(helix.get("length"))
            helix_list = [(i_outter + nt, j_outter - nt) for nt in range(length)]
            basepairs.extend(helix_list)
        # store attributes
        pairs = [0] * len(sequence)
        for i, j in basepairs:
            pairs[i - 1] = j
            pairs[j - 1] = i
        return pd.DataFrame(
            {
                "Nucleotide": np.arange(len(sequence)) + 1,
                "Sequence": list(sequence),
                "Pair": pairs,
                "X_coordinate": xcoords,
                "Y_coordinate": ycoords,
            }
        )

    def read_cte(self):
        """Generates SecondaryStructure object data from a CTE file, including
        nucleotide x and y coordinates.
        """
        # Parse file and get sequence, xcoords, ycoords, and list of pairs.
        return pd.read_table(
            self.filepath,
            sep=r"\s+",
            usecols=[0, 1, 4, 8, 10],
            names=["Nucleotide", "Sequence", "Pair", "X_coordinate", "Y_coordinate"],
            header=0,
        )

    def read_nsd(self):
        """Generates SecondaryStructure object data from an NSD file
        (format for RNAStructure StructureEditor), including nucleotide x and
        y coordinates.
        """
        # Parse file and get sequence, xcoords, ycoords, and list of pairs.
        basepairs, sequence, xcoords, ycoords = [], "", [], []
        with open(self.filepath, "r") as file:
            item = ""
            for line in file.readlines():
                line = line.strip().split(" ")
                if "Strand:[" in line:
                    item = "strand"
                    continue
                elif "]" in line:
                    item = ""
                elif "Pairs:[" in line:
                    item = "pairs"
                    continue
                if item == "strand":
                    nt = {}
                    for field in line:
                        if ":" in field:
                            key, value = field.split(":")
                            nt[key] = value
                    sequence += nt["Base"]
                    xcoords.append(float(nt["X"]))
                    ycoords.append(-float(nt["Y"]))
                elif item == "pairs":
                    for field in line:
                        if field.startswith("Pair"):
                            field = field.strip('Pair:"').split(":")
                            basepairs.append([int(nuc) for nuc in field])
        # store attributes
        pairs = [0] * len(sequence)
        for i, j in basepairs:
            pairs[i - 1] = j
            pairs[j - 1] = i
        return pd.DataFrame(
            {
                "Nucleotide": np.arange(len(sequence)) + 1,
                "Sequence": list(sequence),
                "Pair": pairs,
                "X_coordinate": xcoords,
                "Y_coordinate": ycoords,
            }
        )

    def read_dotbracket(self):
        """Generates SecondaryStructure object data from a dot-bracket notation
        file, including nucleotide x and y coordinates.
        """
        header, seq, bp_str, = (
            "",
            "",
            "",
        )
        with open(self.filepath) as f:
            for line in f:
                if line[0] in ">#":
                    header += line
                    continue
                if all(nt.upper() in "GUACT" for nt in line.strip()):
                    seq += line.strip()
                elif all(s in ".][)(}{><" for s in line.strip()):
                    bp_str += line.strip()
        assert len(bp_str) == len(seq), "db: seq and bp strings are mismatched"
        num = list(range(1, len(seq) + 1))
        bp = [0 for _ in num]
        opens = {"[": [], "{": [], "(": [], "<": []}
        sym_pair = {"]": "[", "}": "{", ")": "(", ">": "<"}
        alphabet = "abcdefghijklmnopqrstuvwxyz"
        sym_pair |= {c: c.upper() for c in alphabet}
        for i, sym in enumerate(bp_str):
            if sym == ".":
                continue
            elif sym in "[{(<" + alphabet.upper():
                opens[sym].append(i)
            elif sym in "]})>" + alphabet:
                j = opens[sym_pair[sym]].pop()
                bp[i] = j + 1
                bp[j] = i + 1
        # store attributes
        return pd.DataFrame(
            {"Nucleotide": np.arange(len(seq)) + 1, "Sequence": list(seq), "Pair": bp}
        )

    def read_r2dt(self):
        """Generates SecondaryStructure object data from an R2DT JSON file,
        including nucleotide x and y coordinates.
        """
        # Parse file and get sequence, xcoords, ycoords, and list of pairs.
        basepairs, sequence, xcoords, ycoords = [], "", [], []
        with open(self.filepath, "r") as file:
            file = json.load(file)
        for nt in file["rnaComplexes"][0]["rnaMolecules"][0]["sequence"]:
            if nt["residueName"] not in ["5'", "3'"]:
                sequence += nt["residueName"]
                xcoords.append(nt["x"])
                ycoords.append(nt["y"])
        sequence = sequence.replace("5'", "").replace("3'", "")
        for bp in file["rnaComplexes"][0]["rnaMolecules"][0]["basePairs"]:
            i = bp["residueIndex1"]
            j = bp["residueIndex2"]
            if [i, j] not in basepairs:
                basepairs.append([i, j])
        # store attributes
        pairs = [0] * len(sequence)
        for i, j in basepairs:
            pairs[i - 1] = j
            pairs[j - 1] = i
        return pd.DataFrame(
            {
                "Nucleotide": np.arange(len(sequence)) + 1,
                "Sequence": list(sequence),
                "Pair": pairs,
                "X_coordinate": xcoords,
                "Y_coordinate": ycoords,
            }
        )

    def read_forna(self):
        """Generates SecondaryStructure object data from a FORNA JSON file,
        including nucleotide x and y coordinates.
        """
        # Parse file and get sequence, xcoords, ycoords, and list of pairs.
        basepairs, sequence, xcoords, ycoords = [], [], [], []
        with open(self.filepath, "r") as file:
            file = json.load(file)
        rna_key = list(file["rnas"].keys())[0]
        rna = file["rnas"][rna_key]
        rna_length = rna["rnaLength"]
        for nt in rna["nodes"][:rna_length]:
            sequence.append(nt["name"])
            xcoords.append(nt["x"])
            ycoords.append(nt["y"])
        for pair in rna["pseudoknotPairs"]:
            basepairs.append(pair)
        for i, j in enumerate(rna["pairtable"][-rna_length:]):
            if (j != 0) and ([j, i + 1] not in basepairs):
                basepairs.append([i + 1, j])
        # store attributes
        pairs = [0] * len(sequence)
        for i, j in basepairs:
            pairs[i - 1] = j
            pairs[j - 1] = i
        return pd.DataFrame(
            {
                "Nucleotide": np.arange(len(sequence)) + 1,
                "Sequence": list(sequence),
                "Pair": pairs,
                "X_coordinate": xcoords,
                "Y_coordinate": ycoords,
            }
        )

    ###########################################################################
    # writing files
    ###########################################################################

    def write_sto(self, outfile, name="seq"):
        """ "write structure file out into Stockholm (STO) file format
        for use for infernal searches"""
        with open(outfile, "w") as out:
            # write header
            out.write("# STOCKHOLM 1.0\n\n")
            namelen = max(len(name), 12)
            out.write(name.ljust(namelen + 1))
            # write nts
            for nt in self.sequence:
                out.write(nt)
            out.write("\n")
            out.write("#=GC SS_cons".ljust(namelen + 1))
            # write dot-brackets
            for nt, pair in enumerate(self.pair_nts):
                if pair == 0:
                    out.write(".")
                elif pair > nt:
                    out.write("(")
                else:
                    out.write(")")
            out.write("\n//\n")

    def write_ct(self, out_file):
        """Writes a ct file from the SecondaryStructure object."""
        with open(out_file, "w") as write_file:
            write_file.write(f"{len(self.nts):6d} {self.filepath}\n")
            for nt, seq, pair in zip(self.nts, self.sequence, self.pair_nts):
                prev = nt - 1
                next = nt + 1 % (self.length + 1)  # last nt goes back to zero
                write_file.write(
                    f"{nt:5d} {seq} {prev:5d} {next:5d} {pair:5d} {nt:5d}\n"
                )

    def write_cte(self, outputPath):
        """writes the current structure out to CTE format for Structure Editor.

        Args:
            outputPath (string): path to output cte file to be created
        """
        # set scaling factors based on data source.
        xscale = {"xrna": 1.525 * 20, "varna": 0.469 * 65, "nsd": 30.5, "cte": 30.5}[
            self.ss_type
        ]
        yscale = {"xrna": -1.525 * 20, "varna": 0.469 * 65, "nsd": 30.5, "cte": 30.5}[
            self.ss_type
        ]
        w = open(outputPath, "w")
        w.write("{0:6d} {1}\n".format(self.length, self.filepath))
        line = "{0:5d} {1} {2:5d} {3:5d} {4:5d} {0:5d} " + ";! X: {5:1f} Y: {6:1f}\n"
        for i in range(self.length):
            xcoord = round(xscale * self.xcoordinates[i])
            ycoord = round(yscale * self.ycoordinates[i])
            num, seq, cti = self.nts[i], self.sequence[i], self.pair_nts[i]
            # no nums after self.length, resets to zero
            nextnum = (num + 1) % (self.length + 1)
            cols = [num, seq, num - 1, nextnum, cti, xcoord, ycoord]
            w.write(line.format(*cols))
        w.close()

    def write_dbn(self, filename, rna_name):
        """Writes dot-bracket notation of current structure to a new file.

        Args:
            filename (str): name of new file
            rna_name (str): name of this RNA for the dbn header
        """
        with open(filename, "w") as write_file:
            write_file.write(f">{rna_name}\n")
            write_file.write(self.sequence + "\n")
            write_file.write(self.get_dotbracket())

    ###########################################################################
    # retrieve structure components or alternative representations
    ###########################################################################

    def get_pairs(self):
        """
        Returns a non-redundant list of base pairs i < j as a array of tuples.
        e.g., [(19,50),(20,49)....]
        """
        out = []
        for left, right in zip(self.nts, self.pair_nts):
            if right != 0 and right > left:
                out.append((left, right))
        return out

    def get_paired_nts(self):
        """Returns a list of residues that are paired."""
        paired_idx = self.pair_nts != 0
        paired_nts = self.data.loc[paired_idx, "Nucleotide"].values.tolist()
        return paired_nts

    def get_unpaired_nts(self):
        """Returns a list of residues that are paired."""
        unpaired_idx = self.pair_nts == 0
        unpaired_nts = self.data.loc[unpaired_idx, "Nucleotide"].values.tolist()
        return unpaired_nts

    def get_junction_nts(self):
        """
        Returns a list of residues at junctions (paired, but adjacent to
        an unpaired residue or the end of a chain)
        """
        junction_residues = []
        for i, nt in enumerate(self.pair_nts):
            if nt == 0:  # if its not paired
                continue
            pi = np.where(self.nts == nt)[0][0]
            neighbors = [i - 1, i + 1, pi - 1, pi + 1]
            neighbors_paired = any(self.pair_nts[neighbors] == 0)
            neighbors_not_in_chain = any([x not in self.nts for x in neighbors])
            if neighbors_paired or neighbors_not_in_chain:
                junction_residues.append(self.nts[i])
        return junction_residues

    def get_nonredundant_ct(self):
        """Returns the ct attribute in non-redundant form - only pairs i<j"""
        pairs = self.get_pairs()
        halfPlexCT = np.zeros_like(self.pair_nts)
        for i, j in pairs:
            halfPlexCT[i - 1] = j
        return halfPlexCT

    def get_helices(self, fill_mismatches=True, split_bulge=True, keep_singles=False):
        """Returns a dictionary of helices from the secondary structure.
        Keys are equivalent to list indices. Values are lists of paired
        nucleotides (1-indexed) in that helix. e.g. {0:[(1,50),(2,49),(3,48)}

        Args:
            fill_mismatches (bool, optional):
                Whether 1-1 and 2-2 bulges are replaced with base pairs.
                Defaults to True.
            split_bulge (bool, optional):
                Whether to split helices on bulges.
                Defaults to True.
            keep_singles (bool, optional):
                Whether to return helices that contain only 1 base-pair.
                Defaults to False.
        """
        # first step is to find all the helices
        rna = self.copy()
        if fill_mismatches:
            rna = rna.fill_mismatches()
        helices = {}
        nt = 0
        heNum = 0
        while nt < len(rna.pair_nts):
            if rna.pair_nts[nt] == 0:
                nt += 1
                continue
            else:
                # skip dups
                if nt > rna.pair_nts[nt]:
                    nt += 1
                    continue
                tempPairs = []
                stillHelix = True
                previous = rna.pair_nts[nt]
                # this is the default behavior which splits helices that are
                # separated by a single nucleotide bulge
                if split_bulge:
                    while stillHelix:
                        # see if the helix juts up against another one
                        if abs(rna.pair_nts[nt] - previous) > 1:
                            break
                        # add the pairs
                        elif rna.pair_nts[nt] != 0:
                            tempPairs.append((nt + 1, rna.pair_nts[nt]))
                            previous = rna.pair_nts[nt]
                            nt += 1
                        else:
                            break
                else:
                    while stillHelix:
                        # Handle bulges on the 3' end of the helix
                        pair = rna.pair_nts[nt]
                        if pair != 0 and not (abs(pair - previous) > 2):
                            tempPairs.append((nt + 1, pair))
                            previous = pair
                            nt += 1
                        # Handle bulges on the 5' end of the helix
                        elif rna.pair_nts[nt + 1] == rna.pair_nts[nt - 1] - 1:
                            nt += 1
                        else:
                            break
                # remove single bp helices
                if len(tempPairs) <= 1 and not keep_singles:
                    continue
                helices[heNum] = tempPairs
                heNum += 1
        return helices

    def extractPK(self, fill_mismatches=True):
        """Returns the pk1 and pk2 pairs from the secondary structure.

        Ignores single base pairs. PK1 is defined as the helix crossing the
        most other bps. if there is a tie, the most 5' helix is called pk1
        returns pk1 and pk2 as a list of base pairs e.g [(1,10),(2,9)...
        """

        def check_overlap(h1, h2):
            # only need to check one set of pairs from each
            # of the helices. Test is to see if they form
            # a cross hatching pattern
            h1_max = max(h1[0])
            h1_min = min(h1[0])
            h2_max = max(h2[0])
            h2_min = min(h2[0])
            if h1_max > h2_min > h1_min and h2_max > h1_max:
                return True
            if h2_max > h1_min > h2_min and h1_max > h2_max:
                return True
            return False

        # make a copy so we don't destroy the original object
        rna = self.copy()
        # get the helices by calling the extract helix function
        helices = rna.get_helices(fill_mismatches=fill_mismatches)
        heNum = len(helices)
        # do the helices overlap? Check for a crosshatching pattern
        # append them to a list if they have it.
        overlaps = []  # stores the helix number
        for i in range(heNum):
            for j in range(i + 1, heNum):
                if check_overlap(helices[i], helices[j]):
                    overlaps.append((i, j))
        # if there are no overlapping bps, return none
        if len(overlaps) == 0:
            return None, None
        # iterate through all pairs of overlapping helices and count
        # the number of pairs in each to determine helix crossing most bps
        crossbps = {}
        for i, j in overlaps:
            if i not in crossbps:
                crossbps[i] = 0
            if j not in crossbps:
                crossbps[j] = 0
            crossbps[i] += len(helices[j])
            crossbps[j] += len(helices[i])
        # convert to list and sort
        crossbps = list(crossbps.items())
        # first sort list by helix number -- i.e. 5' to 3'
        crossbps.sort(key=lambda x: x[0])
        # now sort by which helix crosses the most BPs. This two-step procedure
        # ensures that in the case of ties, we always choose 5'-most helix
        crossbps.sort(key=lambda x: x[1], reverse=True)
        pk1 = helices[crossbps[0][0]]
        pk2 = []
        for k, v in crossbps[1:]:
            pk2.extend(helices[k])
        return pk1, pk2

    def get_dotbracket(self):
        """Returns a dotbracket notation representation of the secondary
        structure.

        Pseudoknot levels:
            1: ()
            2: []
            3: {}
            4: <>
            5: Aa
            6: Bb
            7: Cc
            etc...
        """
        dbn = ["."] * self.length
        pair_list = self.get_pairs()
        pair_levels = np.zeros(len(pair_list), dtype=int)
        levels = ["()", "[]", "{}", "<>"]

        # check for pseudoknots
        level = 0
        while level in pair_levels:
            indexes = np.where(pair_levels == level)[0]
            for i in indexes:
                for j in indexes[indexes > i]:
                    left1, right1 = pair_list[i]
                    left2, right2 = pair_list[j]
                    if pair_levels[i] != pair_levels[j]:
                        continue
                    elif (left1 < left2 < right1) and (right1 < right2):
                        pair_levels[j] += 1
            level += 1

        for pair, level in zip(pair_list, pair_levels):
            dbn[pair[0] - 1] = levels[level][0]
            dbn[pair[1] - 1] = levels[level][1]
        dbn = "".join(dbn)
        return dbn

    def get_human_dotbracket(self):
        """Returns dotbracket notation string representing SecondaryStructure
        object. This is an experimental format designed to be more human
        readable, i.e. no counting of brackets required.


        1)  Letters, instead of brackets, are used to denote nested base pairs.
        2)  Each helix is assigned a letter, which is incremented one letter
            alphabetically from the nearest enclosing stem.
        3)  Non-nested helices (pseudoknots) are assigned canonical brackets.

        From this canonical dbn string:
            how many bases are in the base stem?
            how many nested helices are there?
            ((((....(((.[[..)))))(((...(((..]].))))))))
        Same question, new format:
            AABB....CCC.[[..cccbbBBB...CCC..]].cccbbbaa

        Pseudoknot levels:
            1: Aa, Bb, Cc, etc.
            2: [], 3: {}, 4: <>
        """
        dbn = list(self.get_dotbracket())
        alphabet = "abcdefghijklmnopqrstuvwxyz"
        closed = {n: 0 for n in alphabet}
        helices = self.get_helices(split_bulge=False, keep_singles=True)
        for pairs in helices.values():
            left_most, right_most = pairs[0]
            if dbn[left_most - 1] not in "()":
                continue
            for n in alphabet:
                if left_most > closed[n]:
                    letter = n
                    closed[n] = right_most
                    break
            for left, right in pairs:
                dbn[left - 1] = letter.capitalize()
                dbn[right - 1] = letter
        dbn = "".join(dbn)
        return dbn

    def get_structure_elements(self):
        """This code is not yet implemented.

        Returns a string with a character for each nucleotide, indicating
        what kind of structure element it is a part of.

        Characters:
            Dangling Ends (E)
            Stems (S)
            Hairpin Loops (H)
            Bulges (B)
            Internal Loops (I)
            MultiLoops (M)
            External Loops (X)
            Pseudoknot (P)
        """

    ###########################################################################
    # edit base pairs
    ###########################################################################

    def add_pairs(self, pairs, break_conflicting_pairs=False):
        """Add base pairs to current secondary structure

        Args:
            pairs (list): 1-indexed list of paired residues.
                e.g. [(1, 20), (2, 19)]
        """
        # get a non-redundant list of non-zero pairs
        pairs = list(set([tuple(sorted(p)) for p in pairs if 0 not in p]))
        # find conflicts within provided pairs
        unique, counts = np.unique(np.ravel(pairs), return_counts=True)
        conflicts = unique[counts > 1]
        conflicts = [[pair for pair in pairs if i in pair] for i in conflicts]
        if len(conflicts) > 0:
            print(
                "There are conflicting pairs:\n\t"
                + "\n\t".join([str(conflict) for conflict in conflicts])
            )
            raise ValueError("Input pairs list cannot contain conflicts")
        # find conflicts in structure and either break existing pair or raise
        conflicts = []
        for i, j in pairs:
            current_i = self.pair_nts[i - 1]
            current_j = self.pair_nts[j - 1]
            if current_i not in [0, j]:
                if break_conflicting_pairs:
                    idx = self.data["Nucleotide"] == current_i
                    self.data.loc[idx, "Pair"] = 0
                else:
                    conflicts.append([(i, j), (i, current_i)])
            if current_j not in [0, i]:
                if break_conflicting_pairs:
                    idx = self.data["Nucleotide"] == current_j
                    self.data.loc[idx, "Pair"] = 0
                else:
                    conflicts.append([(i, j), (current_j, j)])
        if len(conflicts) > 0:
            print(
                "There are conflicting pairs:"
                + "\n\t".join([str(conflict) for conflict in conflicts])
                + "use break_conflicting_pairs=True to remove existing pairs."
            )
            raise ValueError(
                "Conflicting pairs found in structure, use "
                "break_conflicting_pairs=True to remove existing pairs."
            )
        # if all is good, update the current structure
        for i, j in pairs:
            self.pair_nts[i - 1] = j
            self.pair_nts[j - 1] = i

    def break_pairs_region(self, start, end, break_crossing=True, inverse=False):
        """Removes pairs from the specified region (1-indexed, inclusive)

        WARNING: this deletes information

        Args:
            start (int): start position
            end (int): end position
            break_crossing (bool, optional): whether to keep pairs that cross
                over the specified region. Defaults to True.
            inverse (bool, optional): invert the behavior. Defaults to False.
        """
        region = pd.Interval(left=start, right=end, closed="both")
        for nt1, nt2 in self.get_pairs():
            pair_idx = self.data["Nucleotide"].isin([nt1, nt2])
            if inverse and (nt1 not in region or nt2 not in region):
                self.data.loc[pair_idx, "Pair"] = 0
            elif not inverse and (nt1 in region or nt2 in region):
                self.data.loc[pair_idx, "Pair"] = 0
            if break_crossing and nt1 < start < end < nt2:
                self.data.loc[pair_idx, "Pair"] = 0

    def fill_mismatches(self, mismatch=1):
        """Adds base pairs to fill 1,1 and optionally 2,2 mismatches.

        mismatch = 1 will fill only 1,1 mismatches
        mismatch = 2 will fill 1,1 and 2,2
        """
        for i in range(len(self.pair_nts) - 3):
            if self.pair_nts[i + 1] == 0:
                if self.pair_nts[i] - self.pair_nts[i + 2] == 2:
                    self.pair_nts[i + 1] = self.pair_nts[i] - 1
            if mismatch == 2 and self.pair_nts[i + 1] + self.pair_nts[i + 2] == 0:
                if self.pair_nts[i] - self.pair_nts[i + 3] == 3:
                    self.pair_nts[i + 1] = self.pair_nts[i] - 1
                    self.pair_nts[i + 2] = self.pair_nts[i] - 2
        return self

    def break_pairs_nts(self, nt_positions):
        """break base pairs at the given list of positions.

        Args:
            nt_positions (list of int): 1-indexed positions
        """
        nt_idx = self.data["Nucleotide"].isin(nt_positions)
        pair_nts = self.data.loc[nt_idx, "Pair"].to_numpy()
        nt_idx &= self.data["Nucleotide"].isin(pair_nts)
        self.data.loc[nt_idx, "Pair"] = 0

    def break_noncanonical_pairs(self):
        """Removes non-canonical basepairs from the secondary structure.

        WARNING: this deletes information.
        """
        seq = self.sequence.upper().replace("T", "U")
        for nt1, nt2 in self.get_pairs():
            bp = "".join([seq[nt1 - 1], seq[nt2 - 1]])
            if bp not in ["AU", "UA", "GC", "CG", "GU", "UG"]:
                mask = self.data["Nucleotide"].isin([nt1, nt2])
                self.data.loc[mask, "Pair"] = 0

    def break_singleton_pairs(self):
        """Removes singleton basepairs from the secondary structure.

        WARNING: This deletes information.
        """
        pair_nts = self.pair_nts
        for nt1, nt2 in self.get_pairs():
            if pair_nts[nt1 - 2] != nt2 - 1 and pair_nts[nt1] != nt2 + 1:
                mask = self.data["Nucleotide"].isin([nt1, nt2])
                self.data.loc[mask, "Pair"] = 0

    ###########################################################################
    # Make calculations based on structures
    ###########################################################################

    def get_distance_matrix(self, recalculate=False):
        """Based on Tom's contact_distance function, but instead returns
        the all pairs shortest paths matrix, and stores it as an attribute. If
        the attribute has already been set, it returns the attribute. This is
        faster than calling contact_distance pairwise to fill the matrix.

        Args:
            recalculate (bool, optional): Set to true to recalculate the matrix
                even if the attribute is set. In case changes to the structure
                have been made.
        """
        if (self.distance_matrix is not None) and not recalculate:
            return self.distance_matrix
        # this method will be used later to make sure a nucleotide hasn't
        # been visited and is within the bounds of the RNA
        def viable(nt):
            not_empty = nt != 0
            valid_so_far = not_empty
            if valid_so_far:
                in_range = min(self.nts) <= nt <= max(self.nts)
                valid_so_far = in_range
            if valid_so_far:
                not_visited = level[np.where(self.nts == nt)[0][0]] == -1
                valid_so_far = not_visited
            return valid_so_far

        # The default value of -1 means that a nt hasn't been visited.
        distance_matrix = np.full((len(self.nts), len(self.nts)), -1)
        for i, nt in enumerate(self.nts):
            # Level keeps track of how far each other nt is from nt
            level = distance_matrix[i, :]
            # create the queue and add nt, set its level as 0
            queue = list()
            queue.append(nt)
            level[i] = 0
            # while there are nucleotides that haven't been visited
            while len(queue) > 0:
                current_nt = queue.pop(0)
                current_i = np.where(self.nts == current_nt)[0][0]
                # Find all the neighbors of the current NT
                NTBack = current_nt - 1
                NTUp = current_nt + 1
                NTPair = self.pair_nts[current_i]
                # check all the neighbors to see if they should be added
                # and if so, record their level(ie their Contact Distance)
                for neighbor in [NTBack, NTUp, NTPair]:
                    if viable(neighbor):
                        queue.append(neighbor)
                        level[np.where(self.nts == neighbor)[0][0]] = (
                            level[current_i] + 1
                        )
            # set row of distance matrix for nt
            distance_matrix[i, :] = level
        # store the distance matrix from the search and return
        return distance_matrix.copy()

    def contact_distance(self, i, j):
        """Returns the contact distance between positions i and j"""
        if self.distance_matrix is None:
            self.distance_matrix = self.get_distance_matrix()
        return self.distance_matrix[i - 1, j - 1]

    def compute_ppv_sens(self, structure2, exact=True):
        """Compute the PPV and sensitivity between self and another
        SecondaryStructure object.

        Args:
            structure2 (SecondaryStructure): The SecondaryStructure to compare to.
            exact (bool, optional): True requires BPs to be exactly correct.
                                    False allows +/-1 bp slippage.
                                    Defaults to True.

        Returns:
            float, float, tuple: sensitivity, PPV, (TP, TP+FP, TP+FN)
        """
        if len(self.pair_nts) != len(structure2.pair_nts):
            raise ValueError("sequence lengths must be the same")
        trues = sum(self.pair_nts != 0) / 2
        positives = sum(structure2.pair_nts != 0) / 2
        true_positives = 0
        for i, v in enumerate(self.pair_nts):
            if v == 0 or v < i:
                continue
            try:
                if v == structure2.pair_nts[i]:
                    true_positives += 1
                elif not exact:
                    neighbors = [
                        structure2.pair_nts[i] - 1,
                        structure2.pair_nts[i] + 1,
                        structure2.pair_nts[i - 1],
                        structure2.pair_nts[i + 1],
                    ]
                    if v in neighbors:
                        true_positives += 1
            except IndexError:
                pass
        try:
            ppv = true_positives / float(positives)
            sens = true_positives / float(trues)
        except ZeroDivisionError:
            ppv, sens = 0.0, 0.0
        return sens, ppv, (true_positives, positives, trues)

    ###########################################################################
    # get other data objects based on this one
    ###########################################################################

    def copy(self):
        return self.get_aligned_data(self.null_alignment)

    def get_aligned_data(self, alignment):
        """Returns a new SecondaryStructure matching the alignment target.

        Args:
            alignment (data.Alignment): an alignment object used to map values
        """
        df = alignment.map_nucleotide_dataframe(self.data)
        df["Pair"] = df["Pair"].fillna(0)
        mask = df["Pair"] != 0
        df.loc[mask, "Pair"] = alignment.map_positions(df.loc[mask, "Pair"].values)
        return SecondaryStructure(
            input_data=df,
            autoscale=False,
            name=self.name,
        )

    def get_interactions_df(self):
        mask = self.data.eval("Pair != 0 & Pair > Nucleotide")
        df = self.data.loc[mask, ["Nucleotide", "Pair"]].copy()
        df["Structure"] = 1
        df = (
            df.rename(columns={"Nucleotide": "i", "Pair": "j"})
            .sort_values(by=["i", "j"])
            .astype({"Structure": "Int32"})
            .reset_index(drop=True)
        )
        return df

    def as_interactions(self, structure2=None):
        """returns list of i, j basepairs as rnavigate.Interactions data.

        Args:
            structure2 (rnavigate.data.SecondaryStructure, optional):
                If provided, basepairs from both structures are included and
                colored by structure (left, right, or both)
                defaults to None.
        """
        if structure2 is None:
            return data.StructureAsInteractions(
                input_data=self,
                sequence=self.sequence,
            )
        elif isinstance(structure2, SecondaryStructure):
            return data.StructureCompareTwo(
                input_data=[self, structure2],
                sequence=self.sequence,
            )
        elif isinstance(structure2, list):
            return data.StructureCompareMany(
                input_data=[self] + structure2,
                sequence=self.sequence,
            )
        raise ValueError(
            "structure2 must be a SecondaryStructure or list of " "SecondaryStructures"
        )

    ###########################################################################
    # edit other attributes
    ###########################################################################

    def normalize_sequence(self, t_or_u="U", uppercase=True):
        super().normalize_sequence(t_or_u=t_or_u, uppercase=uppercase)
        self.data["Sequence"] = list(self.sequence)

    def transform_coordinates(
        self, flip=None, scale=None, center=None, rotate_degrees=None
    ):
        """Perform transformation of structure coordinates

        Args:
            flip (str, optional):
                "horizontal" or "vertical".
                Defaults to None.
            scale (float, optional):
                new median distance of basepairs.
                Defaults to None.
            center (tuple of floats, optional):
                new center x and y coordinate.
                Defaults to None.
            rotate_degrees (float, optional):
                number of degrees to rotate structure.
                Defaults to None.
        """
        structure = StructureCoordinates(
            self.xcoordinates, self.ycoordinates, self.get_pairs()
        )
        if flip is not None:
            structure.flip(flip == "horizontal")
        if scale is not None:
            structure.scale(scale)
        if center is not None:
            structure.center(*center)
        if rotate_degrees is not None:
            structure.rotate(rotate_degrees)
        self.xcoordinates = structure.x
        self.ycoordinates = structure.y


###############################################################################
# end of SecondaryStructure class
###############################################################################


class StructureCoordinates:
    """Helper class to perform structure coordinate transformations"""

    def __init__(self, x, y, pairs=None):
        """initialize structure with arrays holding x and y coordinates. A list
        of base-paired positions is required only if scaling coordinates.

        Args:
            x (numpy.array): x coordinates
            y (numpy.array): y coordinates
            pairs (list of pairs, optional): list of base-paired positions.
        """
        self.x = x
        self.y = y
        self.pairs = pairs

    def get_center_point(self):
        """return the x, y coordinates for the center of structure

        Returns:
            tuple: x and y coordinates of structure center as floats
        """
        x_center = (max(self.x) + min(self.x)) / 2
        y_center = (max(self.y) + min(self.y)) / 2
        return (x_center, y_center)

    def scale(self, median_bp_distance=1):
        """Scale structure such that median base-pair distance is constant.

        Args:
            median_bp_distance (int, optional):
                New median distance between all base-paired nucleotides.
                Defaults to 1.
        """
        bp_distances = []
        for bp in self.pairs:
            y_dist = self.y[bp[0] - 1] - self.y[bp[1] - 1]
            x_dist = self.x[bp[0] - 1] - self.x[bp[1] - 1]
            bp_distances.append((y_dist**2 + x_dist**2) ** 0.5)
        scale_factor = np.median(bp_distances)
        self.y *= median_bp_distance / scale_factor
        self.x *= median_bp_distance / scale_factor

    def flip(self, horizontal=True):
        """Flip structure vertically or horizontally

        Args:
            horizontal (bool, optional):
                whether to flip structure horizontally, otherwise vertically
                Defaults to True.
        """
        if horizontal:
            self.x *= -1
        else:
            self.y *= -1

    def center(self, x=0, y=0):
        """Center structure on given x, y coordinate

        Args:
            x (int, optional): x coordinate of structure center. Defaults to 0.
            y (int, optional): y coordinate of structure center. Defaults to 0.
        """
        x_center, y_center = self.get_center_point()
        self.x -= x_center + x
        self.y -= y_center + y

    def rotate(self, degrees):
        """Rotate structure on current center point

        Args:
            degrees (float): number of degrees to rotate structure
        """
        radians = math.radians(degrees)
        center_x, center_y = self.get_center_point()
        # Translate the coordinates to the origin
        translated_x = self.x - center_x
        translated_y = self.y - center_y
        # Perform rotation
        self.x = translated_x * math.cos(radians) - translated_y * math.sin(radians)
        self.y = translated_x * math.sin(radians) + translated_y * math.cos(radians)
        # Translate the coordinates back to the original center
        self.x += center_x
        self.y += center_y


class SequenceCircle(SecondaryStructure):
    def __init__(self, input_data, gap=30, name=None, **kwargs):
        length = input_data.length
        df = pd.DataFrame(
            {
                "Sequence": list(input_data.sequence),
                "Nucleotide": np.arange(length) + 1,
                "Pair": np.zeros(length),
            }
        )
        gap = 2 * np.pi * gap / 360
        self.gap = gap
        theta_between = (2 * np.pi - self.gap) * 1 / (length - 1)
        self.radius = (length + gap / theta_between) / np.pi
        theta_start = gap / 2
        theta = theta_between * np.arange(length) + theta_start
        df["Theta"] = theta
        super().__init__(
            input_data=df, extension=None, autoscale=False, name=name, **kwargs
        )



"""This submodule defines the Transcriptome and Transcript classes.

Transcriptome objects require genome fasta and annotation gtf files. Then, if
provided with a transcript ID, it will return a Transcript object.

Transcript objects contain a transcript sequence and genome coordinates.
It can return annotations for CDS, UTR, and exon junctions.
It can be used with Bed objects to convert genome coordinate data to transcript
coordinate RNAvigate data classes.
"""


import pandas as pd
import numpy as np
from Bio import SeqIO
from rnavigate import data
from pathlib import Path


data_path = Path(__file__).parent.parent.parent / "reference_data"
human_chromosome_ids = {
    "chr1": "NC_000001",
    "chr2": "NC_000002",
    "chr3": "NC_000003",
    "chr4": "NC_000004",
    "chr5": "NC_000005",
    "chr6": "NC_000006",
    "chr7": "NC_000007",
    "chr8": "NC_000008",
    "chr9": "NC_000009",
    "chr10": "NC_000010",
    "chr11": "NC_000011",
    "chr12": "NC_000012",
    "chr13": "NC_000013",
    "chr14": "NC_000014",
    "chr15": "NC_000015",
    "chr16": "NC_000016",
    "chr17": "NC_000017",
    "chr18": "NC_000018",
    "chr19": "NC_000019",
    "chr20": "NC_000020",
    "chr21": "NC_000021",
    "chr22": "NC_000022",
    "chrX": "NC_000023",
    "chrY": "NC_000024",
    "chrMT": "NC_012920",
}


class Transcript(data.Sequence):
    def __init__(
        self,
        parent,
        name,
        sequence,
        chromosome,
        strand,
        coordinates,
        tx_info,
        cds_coors=None,
        other_features=None,
    ):
        super().__init__(sequence)
        self.name = name
        self.parent = parent
        self.chromosome = chromosome
        self.strand = strand
        self.coordinates = coordinates
        self.tx_info = tx_info
        self.cds_coors = cds_coors
        self.other_features = other_features
        self.coordinate_df = self.get_coordinate_df()

    def get_coordinate_df(self):
        coordinates = [np.arange(x, y + 1) for x, y in zip(*self.coordinates)]
        coordinates = np.concatenate(coordinates)
        if self.strand == "-":
            coordinates = coordinates[::-1]
        df = pd.DataFrame(
            {
                "Sequence": list(self.sequence),
                "Nucleotide": np.arange(self.length) + 1,
                "Coordinate": coordinates,
            }
        )
        return df

    def get_tx_coordinate(self, coordinate):
        df = self.coordinate_df
        coordinate = df["Coordinate"] == coordinate
        if sum(coordinate) == 0:
            return None
        return df.loc[coordinate, "Nucleotide"].values[0]

    def get_tx_range(self, start, stop):
        df = self.coordinate_df
        nts = df.loc[df["Coordinate"].between(start, stop), "Nucleotide"]
        return [nts.min(), nts.max()]

    def get_cds_annotation(self, **kwargs):
        cds = self.get_tx_range(*sorted(self.cds_coors))
        return data.Annotation(
            input_data=[cds],
            annotation_type="spans",
            sequence=self.sequence,
            name="CDS",
            **kwargs,
        )

    def get_cds_domains(self):
        cds = self.get_tx_range(*sorted(self.cds_coors))
        spans = [[1, cds[0] - 1], cds, [cds[1] + 1, self.length]]
        data.domains(
            input_data=spans,
            names=["5' UTR", "CDS", "3' UTR"],
            colors=["skyblue", "palegreen", "orchid"],
            sequence=self.sequence,
        )

    def get_junctions_annotation(self, **kwargs):
        junctions = [self.coordinates[0][1:], self.coordinates[1][:-1]]
        junctions = [
            self.get_tx_range(*sorted(junction)) for junction in zip(*junctions)
        ]
        return data.Annotation(
            input_data=junctions,
            annotation_type="spans",
            sequence=self.sequence,
            name="exon junctions",
            **kwargs,
        )

    def get_exon_annotation(self, exon_number, **kwargs):
        if self.strand == "-":
            exon_number = len(self.coordinates[0]) + 1 - exon_number
        try:
            exon = [
                self.coordinates[0][exon_number - 1],
                self.coordinates[1][exon_number - 1],
            ]
        except IndexError as e:
            raise ValueError(f"exon number {exon_number} is too high") from e
        exon = self.get_tx_range(*sorted(exon))
        return data.Annotation(
            input_data=[exon],
            annotation_type="spans",
            sequence=self.sequence,
            name=f"exon {exon_number}",
            **kwargs,
        )

    def get_exon_domains(self):
        exons = [self.get_tx_range(*sorted(e)) for e in zip(*self.coordinates)]
        data.domains(
            input_data=exons,
            names=[f"exon {i}" for i in range(len(exons))],
            colors=[["lightsalmon", "gold"][i % 2] for i in range(len(exons))],
            sequence=self.sequence,
        )


class Transcriptome:
    def __init__(self, genome, annotation, path=data_path, chr_ids=None):
        if path is not data_path:
            path = Path(path)
        if chr_ids is None:
            chr_ids = human_chromosome_ids
        self.genome = data_path / genome
        self.annotation = data_path / annotation
        self.chr_ids = chr_ids

    def get_transcript(self, transcript_id):
        return self.get_transcripts([transcript_id])[transcript_id]

    def get_transcripts(self, transcript_ids):
        def parse_entry(entry):
            entry = entry.strip().split("\t")
            if entry[2] == "transcript":
                info = entry[8].split(";")
                info = [item.strip().split(" ") for item in info[:-1]]
                info = {k: v.strip('"') for k, v in info}
            else:
                info = {}
            return {
                "chromosome": entry[0],
                "db": entry[1],
                "feature": entry[2],
                "start": entry[3],
                "stop": entry[4],
                "strand": entry[6],
                "info": info,
            }

        with open(self.annotation) as txome:
            tx_info, chromosome, strand = {}, {}, {}
            tx_lists = {tx_id: [] for tx_id in transcript_ids}
            start_coors = tx_lists.copy()
            stop_coors = tx_lists.copy()
            cds_coors = tx_lists.copy()
            other_features = tx_lists.copy()
            for entry in txome.readlines():
                tx_id = None
                for this_id in transcript_ids:
                    if f'transcript_id "{this_id}"' in entry:
                        tx_id = this_id
                if tx_id is None:
                    continue
                tx = parse_entry(entry)
                if tx["feature"] == "transcript":
                    tx_info[tx_id] = tx["info"]
                    chromosome[tx_id] = tx["chromosome"]
                    strand[tx_id] = tx["strand"]
                elif tx["feature"] == "exon":
                    start_coors[tx_id].append(int(tx["start"]))
                    stop_coors[tx_id].append(int(tx["stop"]))
                elif tx["feature"] in ["start_codon", "stop_codon"]:
                    cds_coors[tx_id].append(int(tx["start"]))
                    cds_coors[tx_id].append(int(tx["stop"]))
                elif tx["feature"] in ["UTR", "CDS"]:
                    continue
                else:
                    other_features[tx_id].append(
                        {key: tx[key] for key in ["feature", "start", "stop"]}
                    )
        transcripts = []
        coordinates = [
            (sorted(start_coors[id]), sorted(stop_coors[id])) for id in transcript_ids
        ]
        sequences = self.get_sequences(chromosome, coordinates, strand)
        for tx_id in transcript_ids:
            cds_coors[tx_id] = [min(cds_coors[tx_id]), max(cds_coors[tx_id])]
            transcripts.append(
                Transcript(
                    parent=self,
                    name=tx_id,
                    sequence=sequences[tx_id],
                    chromosome=chromosome[tx_id],
                    strand=strand[tx_id],
                    coordinates=coordinates,
                    tx_info=tx_info[tx_id],
                    cds_coors=cds_coors[tx_id],
                    other_features=other_features[tx_id],
                )
            )
        return transcripts

    def get_sequences(self, chromosomes, coordinates, strands):
        sequences = {tx_id: "" for tx_id in chromosomes}
        for record in SeqIO.parse(self.genome, "fasta"):
            for tx_id, chromosome in chromosomes.items():
                if record.id.startswith(self.chr_ids[chromosome]):
                    for start, end in zip(*coordinates[tx_id]):
                        sequences[tx_id] += str(record.seq[start - 1 : end])
        for tx_id, seq in sequences.values():
            seq = seq.upper().replace("T", "U")
            # reverse compliment if on minus strand
            if strands[tx_id] == "-":
                seq = "".join(["AUCG"["UAGC".index(nt)] for nt in seq[::-1]])
            sequences[tx_id] = seq
        return sequences

    def get_sequence(self, chromosome, coordinates, strand):
        seq = ""
        chromosome = self.chr_ids[chromosome]
        for record in SeqIO.parse(self.genome, "fasta"):
            if record.id.startswith(chromosome):
                for start, end in zip(*coordinates):
                    seq += str(record.seq[start - 1 : end])
        seq = seq.upper().replace("T", "U")
        # reverse compliment if on minus strand
        if strand == "-":
            seq = "".join(["AUCG"["UAGC".index(nt)] for nt in seq[::-1]])
        return seq



from rnavigate import plots
import numpy as np


class SS(plots.Plot):
    """Plot secondary structure diagrams.

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.
    **kwargs
        Keyword arguments passed to `rnavigate.plots.Plot`.

    Attributes
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    xlims : list of float
        x limits of the plot.
    ylims : list of float
        y limits of the plot.
    i : int
        Index of the current plot.
    """

    def __init__(self, num_samples, **kwargs):
        """Initialize the plot."""
        super().__init__(num_samples, sharey=True, sharex=True, **kwargs)
        for i in range(self.length):
            ax = self.get_ax(i)
            ax.axis("off")
        self.xlims = [-2, 2]
        self.ylims = [-2, 2]

    def set_figure_size(
        self,
        height_ax_rel=0.2,
        width_ax_rel=0.2,
        width_ax_in=None,
        height_ax_in=None,
        height_gap_in=0.5,
        width_gap_in=0.2,
        top_in=1,
        bottom_in=0.5,
        left_in=0.5,
        right_in=0.5,
    ):
        """Set the figure size.

        Parameters
        ----------
        height_ax_rel : float, defaults to 0.2
            Height of the axes relative to the y-axis limits.
        width_ax_rel : float, defaults to 0.2
            Width of the axes relative to the x-axis limits.
        width_ax_in : float
            Width of the axes in inches.
        height_ax_in : float
            Height of the axes in inches.
        height_gap_in : float, defaults to 0.5
            Height of the gap between axes in inches.
        width_gap_in : float, defaults to 0.2
            Width of the gap between axes in inches.
        top_in : float, defaults to 1
            Height of the top margin in inches.
        bottom_in : float, defaults to 0.5
            Height of the bottom margin in inches.
        left_in : float, defaults to 0.5
            Width of the left margin in inches.
        right_in : float, defaults to 0.5
            Width of the right margin in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )

    def plot_data(
        self,
        structure,
        interactions=None,
        interactions2=None,
        profile=None,
        annotations=None,
        label="",
        colors=None,
        nt_ticks=None,
        bp_style="dotted",
    ):
        """Plot the data on the current axes.

        Parameters
        ----------
        structure : rnavigate.data.SecondaryStructure
            Structure object with diagram drawing coordinates.
        interactions : rnavigate.data.Interactions, optional
            Interactions object to plot as lines between nucleotides.
        interactions2 : rnavigate.data.Interactions, optional
            Interactions object to plot as lines between nucleotides.
        profile : rnavigate.data.Profile, optional
            Profile object used to color nucleotides.
        annotations : list of rnavigate.data.Annotation, optional
            Annotation objects to highlight regions or nucleotides of interest.
        label : str, optional
            Label for the plot title.
        colors : dict, optional
            Dictionary of colors for each plot element. Keys are "sequence",
            "nucleotides", "structure", and "basepairs". Values are either
            matplotlib colors or strings specifying the color scheme.
        nt_ticks : int, optional
            Number of nucleotides between tick marks.
        bp_style : "dotted", "solid", or "conventional", defaults to "dotted"
            Style of base pair lines.
        """
        if annotations is None:
            annotations = []
        ax = self.get_ax()
        if colors is None:
            colors = {}
        colors = {
            "sequence": None,
            "nucleotides": "sequence",
            "structure": "grey",
            "basepairs": "grey",
        } | colors
        for key in ["nucleotides", "structure", "basepairs", "sequence"]:
            if isinstance(colors[key], str) and colors[key] == "contrast":
                continue
            elif colors[key] is None:
                continue
            colors[key], colormap = structure.get_colors(
                colors[key],
                profile=profile,
                structure=structure,
                annotations=annotations,
            )
            self.add_colorbar_args(colormap)
        if isinstance(colors["sequence"], np.ndarray):
            colors["nucleotides"], _ = structure.get_colors("white")
        elif colors["sequence"] == "contrast":
            colors["sequence"] = plots.get_contrasting_colors(colors["nucleotides"])
        if colors["sequence"] is not None:
            plots.plot_sequence_ss(ax, structure, colors["sequence"])
        if colors["nucleotides"] is not None:
            plots.plot_nucleotides_ss(ax, structure, colors["nucleotides"])
        if colors["structure"] is not None:
            plots.plot_structure_ss(ax, structure, colors["structure"])
        if colors["basepairs"] is not None:
            plots.plot_basepairs_ss(ax, structure, bp_style)
        if nt_ticks is not None:
            plots.plot_positions_ss(ax, structure, nt_ticks)
        if interactions is not None:
            plots.plot_interactions_ss(ax, structure, interactions)
            self.add_colorbar_args(interactions.cmap)
        if interactions2 is not None:
            plots.plot_interactions_ss(ax, structure, interactions2)
            self.add_colorbar_args(interactions2.cmap)
        for annotation in annotations:
            plots.plot_annotation_ss(ax, structure, annotation)
        ax.set_title(label)
        x = structure.xcoordinates
        y = structure.ycoordinates
        xlims = ax.get_xlim()
        ylims = ax.get_ylim()
        xlims = [min([xlims[0], min(x) - 2]), max([xlims[1], max(x) + 2])]
        ylims = [min([ylims[0], min(y) - 2]), max([ylims[1], max(y) + 2])]
        ax.set(ylim=ylims, xlim=xlims)
        self.i += 1



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Module for generating secondary and tertiary structure figures
# Seth Veenbaas
# Weeks Lab, UNC-CH
# 2022
#
# Version 1.3.0
#
# -----------------------------------------------------------------------------
import os
import pickle
import ast
from glob import glob
from pylab import *
from prody import *
import numpy as np
import pandas as pd
import seaborn as sns
import rnavigate as rnav
from fpocketR import make3D
from pymol import cmd
from matplotlib.colors import LinearSegmentedColormap


def make_figures(
        pdb : str,
        state : int,
        pc_df : pd.DataFrame,
        rna_coords : prody.AtomGroup,
        ss :str ,
        analysis : str,
        name : str,
        chain : str,
        dpi : int,
        zoom : int,
        offset : int,
        connectpocket : bool,
        alignligand : str,
    ) -> dict:

    # Get the rna sequnece length from the .pdb file.    
    pdb_seq_len = len(rna_coords.select("name O2'"))

    # Gets secondary structure input file required for making figures.
    try:
        if os.path.isfile(ss) is False:
            raise Exception

    except TypeError:
        ss = False

        # If not making 2D figures than get rna sequence length from .pdb file.
        rna_seq_len = pdb_seq_len
        pass

    except Exception:
        print('\n ERROR: An valid .nsd file is required to make 2D figures.\n'
              'We suggest generating secondary structure files by:\n'
              '1) Converting .pdb to .ct files at:\n'
              'http://rnapdbee.cs.put.poznan.pl\n'
              '2) Making .ss from .ct with StructureEditor avalible at:\n'
              'https://rna.urmc.rochester.edu/RNAstructureDownload.html\n')
        ss = False

        # If not making 2D figures than get rna sequence length from .pdb file.
        rna_seq_len = pdb_seq_len
        pass

    else:
        # Makes RNAvigate object for rna secondary structure.
        rna_map = rnav.Sample(sample=name, ss=ss)

        # Gets length of RNA. ss_seq_len needed to make figures.
        ss_seq_len = rna_map.data["ss"].length

        if pdb_seq_len != ss_seq_len:
            print(f'WARNING: The PDB sequence length ({pdb_seq_len}) is different '
                  f'than the ss seqence length ({ss_seq_len}).\n'
                  f'Make sure these files are correct.\n')

        # Define RNA sequence length based on ss file.
        rna_seq_len = ss_seq_len
    
    print_pockets(pc_df)

    pc_df.to_csv(f'{analysis}/{name}_out_pocket_characteristics.csv',
                 index=False, float_format='%.2g')

    # Get color maps to color 2D and 3D figures
    seq_cmap, pocket_cmap, pocket_nt_color = get_colorNT(
        pc_df, rna_seq_len, offset, rna_coords, ss, chain)
    
    if state:
        state_pocket_cmap = {int(state): pocket_cmap}
        state_pocket_nt_color = {int(state): pocket_nt_color}
        # Save pocket_cmap to .pkl file for multistate figure generation
        with open(f'{analysis}/{name}_maps.pkl', "wb") as file:
            pickle.dump([state_pocket_cmap, state_pocket_nt_color], file)

    # Get 2D figures based on secondary structure in .ss file.
    if ss:
        make_2D_figure(ss, seq_cmap, pocket_nt_color,
                      analysis, name, connectpocket)

    # Get 3D PyMol figures based on the real_sphere.pdb file.
    make_3D_figure(pdb, state, analysis, name, dpi,
                  chain, zoom, pocket_cmap, alignligand)
    
    return pocket_cmap

# -----------------------------------------------------------------------------
def print_pockets(pc_df : pd.DataFrame):
    pc_passing = pc_df[pc_df['Filter'] == 'Pass'].copy()
    if 1 in pc_passing["Pocket"].to_list():
        print(f"\nTotal pockets: {(pc_passing['Filter'] == 'Pass').sum()}")
    else: 
        print('\nTotal pockets: 0')
    if 'Known' in pc_passing["Type"].to_list():
        print(
            f"Known pockets: {pc_passing['Type'].value_counts()['Known']}\n")
    else:
        print('Known pockets: 0\n')

def get_pdb_offset(
        rna_coords : prody.AtomGroup,
        nt : int,
        offset : int,
        chain : str
    ) -> int:
    """Used for for calculating nucleotide offsets for RNA with 2 chains.

    Args:
        rna_coords (object): Prody atom group of RNA coordinates.
        nt (int): Index of the nucleotide of interest for calculating an offset
        offset (int): Sequence offset between .pdb and .ss file (1st chain).
        chain (str): Chain identifier for RNA chain.

    Returns:
        int: 2D structure index for the specified nucleotide
    """
    second_chain = str(chain.split(",")[1])
    offset2NT = rna_coords.select(
        f'chain {second_chain}').copy().getResnums()[0]
    first_chain = str(chain.split(",")[0])
    offsetNT = rna_coords.select(
        f'chain {first_chain}').copy().getResnums()[-1]
    offset2 = offset2NT - offsetNT - 1

    if nt < offset2NT:
        cmap_idx = nt - 1 - offset
    elif nt >= offset2NT:
        cmap_idx = nt - 1 - offset - offset2
        
    return cmap_idx


def get_colorNT(
        pc_df : pd.DataFrame,
        rna_seq_len : int,
        offset : int,
        rna_coords : prody.AtomGroup,
        ss : str,
        chain : str,
    ) -> tuple[list[tuple], list[tuple], list[dict]]:
    """Generates a color map of pocket loations throughout the RNA.

    Args:
        pc_df (dataframe): Contains all the characteristics for each pocket.
        rna_seq_len (int): Number of nucleotides in RNA sequence.
        offset (int): Sequence offset between .pdb and .ss file (1st chain).
        rna_coords (object): Prody atom group of RNA coordinates.
        ss (str): Path to input secondary structure drawing.
        chain (str): Chain identifier for RNA chain.

    Returns:
        list[tuple]: Per nucleotide color map. Index = NT sequence. Value = color.
        list[tuple]: Per pocket color map. Index = pocket residue number. Value = color.
        list[dict]: Per site nucleotides and color for RNAvigate annotations.
    """

    seq_cmap = [(1, 1, 1, 1)] * rna_seq_len
    pocket_cmap = {}
    pocket_nt_color = []

    # Cubehelix color map for known nucleotides/pockets.
    known_cmap = sns.color_palette(
        'ch: 1.1, rot=0.1, gamma=0.6, light=0.65, dark=0.35, hue=3, reverse=1',
        as_cmap=True)

    # Cubehelix color map for novel nucleotides/pockets.
    novel_cmap = sns.color_palette(
        'ch: 0.09, rot=2.55, gamma=0.85, light=0.45, dark=0.25, hue=1.25, reverse=1',
        as_cmap=True)

    # Creates DataFrame containing only pockets that Pass the quality filter.
    pocket_df = pc_df[pc_df['Filter'] == 'Pass'].copy()
    if len(pocket_df) is None:
        return None, None

    if pocket_df[pocket_df['Type'] == 'Known'] is None:
        known_len = 0
    else:
        known_len = len(pocket_df[pocket_df['Type'] == 'Known'])

    if pocket_df[pocket_df['Type'] == 'Novel'] is None:
        novel_len = 0
    else:
        novel_len = len(pocket_df[pocket_df['Type'] == 'Novel'])

    cmaps = {'Known': known_cmap, 'Novel': novel_cmap}
    lengths = {'Known': known_len - 1, 'Novel': novel_len - 1}
    counts = {'Known': 0, 'Novel': 0}

    colored_nt_list = []
    known_colored_nt_list = []

    for idx, row in pocket_df.iterrows():
        pocketNT = row['PocketNT']
        poc_type = str(row['Type'])
        poc_num = int(row['Pocket'])

        # Gets equally spaced colors for each pocket from color map.
        if lengths[poc_type] == 0:
            color = cmaps[poc_type](0)
        else:
            color = cmaps[poc_type](counts[poc_type] / lengths[poc_type])

        # Assigns color to each pocket.
        pocket_cmap[poc_num] = color
        counts[poc_type] += 1

        # Colors nucleotides with matching color to their associated pocket.
        # If a single nucleotide is in contact with multiple pockets:
        # -Priority is given to higher ranked pockets.
        # -Color of a known pocket will not be overwritten.
        cmap_idx_per_pocket = []
        for nt in pocketNT:
            # Offset used to match index of .pdb input to .ss output.
            if len(chain) == 1:
                cmap_idx = nt - 1 - offset

            # If the RNA contains 2 chains get the second offset.
            elif len(chain.split(',')) == 2:
                cmap_idx = get_pdb_offset(rna_coords, nt, offset, chain)
            else:
                print('WARNING: 2D figures can only be made from RNA strctures \n'
                      'composed of a maxiumum of 2 chains (-c).')
                ss = None
                break

            if ss:
                if nt not in colored_nt_list and \
                        nt not in known_colored_nt_list:
                    seq_cmap[cmap_idx] = color
                elif poc_type == 'Known' and \
                        nt not in known_colored_nt_list:
                    seq_cmap[cmap_idx] = color

            colored_nt_list.append(nt)
            if poc_type == 'Known':
                known_colored_nt_list.append(nt)

            cmap_idx_per_pocket.append(int(cmap_idx + 1))
        pocket_nt_color.append(
            {'pocket': f'pocket {poc_num}','nucleotides': cmap_idx_per_pocket, 'color': color})

    return seq_cmap, pocket_cmap, pocket_nt_color


def make_2D_figure(
        ss : str,
        seq_cmap : list[tuple],
        pocket_nt_color : list[dict],
        analysis : str,
        name : str,
        connectpocket : bool,
    ) -> None:
    """Uses RNAvigate to plot pocketNT onto the RNA secondary structure and
       saves resulting 2D figure as png and svg.

    Args:
        ss (str): Path to input secondary structure drawing.
        pdb_ligand (str): Path to input .pdb file.
        chain (str): Indicated desired PDB chain (default='A').
        offset (int): Sequence offset between .pdb and .ss file (default=0).
        seq_cmap list(tuple): RGB color codes for each nucleotide.
        analysis (str): Path to the analysis output directory.
        name (str): Output file name prefix (default=pdb_name).
        connectpocket (boolean): Connects pockets in 2D figure (Default=False).
    """
    print('Making 2D figure.\n')
    rna_map = rnav.Sample(sample=name, ss=ss)

    rnav.styles.settings["ss"]["nucleotides"]["s"] = 7**2
    # Makes 2D figures with transparent colored lines
    # that connect all the nucleotides associated with a pocket.
    if connectpocket:
        pocket_annotations = []
        for d in pocket_nt_color:            
            pocket_annotations.append(d['pocket'])
            rna_map.set_data(
                f'{d["pocket"]}', 
                {
                    'group': d['nucleotides'],
                    'sequence': 'ss',
                    'color': d['color'],
                    'name': d['pocket']
                }
                )

        rnav.plot_ss(
            samples=[rna_map],
            structure='ss',
            annotations=pocket_annotations,
            colors={
                "sequence": "contrast",
                "nucleotides": seq_cmap
                }, 
            bp_style="line"
            )

    # Makes 2D figures without connecting nucelotides.
    else:
        rnav.plot_ss(
            samples=[rna_map],
            structure='ss',
            colors={
                "sequence": "contrast",
                "nucleotides": seq_cmap,
                },
            bp_style="line"
            )

    # Saves 2D figure was .png file.
    plt.savefig(f'{analysis}/{name}_2D.png', dpi=300, format=None,
                metadata=None, bbox_inches=None, pad_inches=0.1,
                facecolor='auto', edgecolor='auto', backend=None)
    # Saves 2D figure as .svg (editable) file.
    plt.savefig(f'{analysis}/{name}_2D.svg', dpi=300, format=None,
                metadata=None, bbox_inches=None, pad_inches=0.1,
                facecolor='auto', edgecolor='auto', backend=None)
    plt.close()


def make_3D_figure(
        pdb : str,
        state : int,
        analysis : str,
        name : str,
        dpi : int,
        chain : str,
        zoom : float,
        pocket_cmap : dict[int, tuple[float, float, float, float]],
        alignligand : str,
    ) -> None:
    """Generates a 3D figure and pymol session file for a single state.

    Args:
        pdb (str): Path to input .pdb file.
        state (int): Structural state to analyze.
        analysis (str): path to directory containing fpocket outputs.
        name (str): Output file name prefix (default={pdb_name}).
        dpi (int): Figure resolution in dpi (dots per linear inch).
        chain (str): Chain identifier for desired RNA chain.
        zoom (float): Zoom buffer distance (Å) for creating 3D figures.
        pocket_cmap list(tuple): Per pocket color map. Index = pocket residue number. Value = color.
        alignligand (str): Align ligand to pymol output (defualt=True).
    """
    print('Making 3D figure.\n')
    real_sphere_name = f'{name}_out_real_sphere'
    real_sphere_pdb = os.path.join(analysis, f'{real_sphere_name}.pdb')
    make3D.load_pdb(real_sphere_pdb)
    if alignligand:
        make3D.alignligand(real_sphere_name, alignligand, state)
    make3D.set_default()
    if pocket_cmap:
        make3D.color_pockets(pocket_cmap)
    make3D.save_3D_figure(analysis, name, dpi, chain, zoom)


def get_all_states_3D_figure(
        num_states : int,
        out : list[str],
        name : str,
        multistate_pocket_cmap : dict,
        alignligand : str,
        dpi : int,
        chain : str,
        zoom : float,
    ) -> None:
    """Generates a 3D figure and pymol session file with all states.

    Args:
        out (str): Name of fpocket output parent directory name.
        name (str): PDB identification code for output .pdb file.
        multistate_pocket_cmap (dict): Per pocket color map. Index = pocket residue number. Value = color.
        dpi (int): Figure resolution in dpi (dots per linear inch).
        chain (str): Chain identifier for desired RNA chain.
        zoom (float): Zoom buffer distance (Å) for creating 3D figures.
    """
    print(f'Making multistate 3D figures.\n')
    for state in range(1, num_states+1):
        real_sphere = f'{out}/{name}_clean_state{state}_out/{name}_state{state}_out_real_sphere.pdb'
        cmd.load(f'{real_sphere}', f'state{state}', partial=1)
        if alignligand:
            make3D.alignligand(f'state{state}', alignligand, target_state=state)
        elif (state > 1):
            cmd.align(f'state{state}', f'state1')
        make3D.color_multistate_pockets(f'state{state}', multistate_pocket_cmap[state])
        make3D.transparent_pocket(f'state{state}', state, multistate_pocket_cmap)

    make3D.set_default()
    make3D.make_multistate()
    cmd.refresh()
    make3D.save_3D_figure(out, f'{name}_all_states', dpi, chain, zoom)
    
    
def count_rna_residues(multistate_pocket_nt_color, rna_length):
    """Counts RNA residue occurrences. Handles various input types."""

    if not isinstance(multistate_pocket_nt_color, dict):
        return {}

    residue_counts = {i: 0 for i in range(1, rna_length + 1)}
    for pocket_data_list in multistate_pocket_nt_color.values():
        if not isinstance(pocket_data_list, list):
            continue  # Skip if not a list

        for pocket_data in pocket_data_list:
            if not isinstance(pocket_data, dict) or 'nucleotides' not in pocket_data:
                continue  # Skip if not a dict or missing 'nucleotides'

            nucleotides = pocket_data['nucleotides']
            if not isinstance(nucleotides, list):
                continue  # Skip if 'nucleotides' is not a list

            for residue in nucleotides:
                if isinstance(residue, int):
                    residue_counts[residue] = residue_counts.get(residue, 0) + 1
    return residue_counts
    
    
def get_all_states_2D_figure(
    name : str,
    out : str,
    ss : str,
    num_states : int,
    multistate_pocket_nt_color : dict[int, dict]
):
    
    # Makes RNAvigate object for rna secondary structure.
    rna_map = rnav.Sample(sample=name, ss=ss)

    # Gets length of RNA. ss_seq_len needed to make figures.
    ss_seq_len = rna_map.data["ss"].length   
    nucleotide_count = count_rna_residues(multistate_pocket_nt_color, ss_seq_len)
    # max_count = max(nucleotide_count.values()) if nucleotide_count.values() else 0  # Avoid divide by zero
    max_count = num_states
    normalized_density = {k: v / max_count for k, v in nucleotide_count.items()}

    # Get the rocket_r colormap
    rocket_r_cmap = sns.color_palette("rocket_r", as_cmap=True)

    # Create a custom colormap that includes white for 0
    colors = [(1, 1, 1)] + [rocket_r_cmap(i) for i in np.linspace(0, 1, 256)]
    custom_cmap = LinearSegmentedColormap.from_list("custom_cmap", colors)

    # Create the seq_cmap list using the normalized_density dictionary
    seq_cmap = [
        custom_cmap(0) if normalized_density[i] == 0 else custom_cmap(normalized_density[i])
        for i in range(1, len(normalized_density) + 1)
    ]

    rnav.styles.settings["ss"]["nucleotides"]["s"] = 7**2

    rnav.plot_ss(
        samples=[rna_map],
        structure='ss',
        colors={
            "sequence": "contrast",
            "nucleotides": seq_cmap,
            },
        bp_style="line",
        colorbars=True,
        )

    # Saves 2D figure was .png file.
    plt.savefig(f'{out}/{name}_2D_pocket_density.png', dpi=300, format=None,
                metadata=None, bbox_inches=None, pad_inches=0.1,
                facecolor='auto', edgecolor='auto', backend=None)
    # Saves 2D figure as .svg (editable) file.
    plt.savefig(f'{out}/{name}_2D_pocket_density.svg', dpi=300, format=None,
                metadata=None, bbox_inches=None, pad_inches=0.1,
                facecolor='auto', edgecolor='auto', backend=None)



"""Circle plot class."""

from rnavigate import plots, data
import numpy as np


class Circle(plots.Plot):
    """Create a circle plot.

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.
    **kwargs
        Keyword arguments passed to :class:`rnavigate.plots.Plot`.

    Attributes
    ----------
    zorder : dict
        Dictionary of zorder values for each plot element.
    fig : matplotlib.figure.Figure
        Figure object.
    ax : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    i : int
        Index of current axes object. Increments with each call to
        :meth:`rnavigate.plots.Circle.plot_data`.
    """

    def __init__(self, num_samples, **kwargs):
        """Initialize Circle object."""
        try:
            kwargs["subplot_kw"].update({"projection": "polar"})
        except KeyError:
            kwargs["subplot_kw"] = {"projection": "polar"}
        super().__init__(num_samples, **kwargs)
        self.zorder = {
            "annotations": 0,
            "data": 5,
            "nucleotide": 10,
            "sequence": 15,
            "position": 20,
        }

    def set_figure_size(
        self,
        height_ax_rel=0.035,
        width_ax_rel=0.035,
        width_ax_in=None,
        height_ax_in=None,
        height_gap_in=1,
        width_gap_in=1,
        top_in=1,
        bottom_in=0.5,
        left_in=0.5,
        right_in=0.5,
    ):
        """Set figure size.

        Parameters
        ----------
        height_ax_rel : float, defaults to 0.035
            Height of axes relative to the y-axis limits.
        width_ax_rel : float, defaults to 0.035
            Width of axes relative to the x-axis limits.
        width_ax_in : float, defaults to None (overridden by width_ax_rel)
            Width of axes in inches.
        height_ax_in : float, defaults to None (overridden by height_ax_rel)
            Height of axes in inches.
        height_gap_in : float, defaults to 1
            Height of gap between axes in inches.
        width_gap_in : float, defaults to 1
            Width of gap between axes in inches.
        top_in : float, defaults to 1
            Height of top margin in inches.
        bottom_in : float, defaults to 0.5
            Height of bottom margin in inches.
        left_in : float, defaults to 0.5
            Width of left margin in inches.
        right_in : float, defaults to 0.5
            Width of right margin in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )

    def plot_data(
        self,
        sequence,
        structure=None,
        structure2=None,
        interactions=None,
        interactions2=None,
        profile=None,
        annotations=None,
        label=None,
        colors=None,
        gap=30,
        nt_ticks=(20, 5),
    ):
        """Plot data on the current (or specified) axes.

        Parameters
        ----------
        sequence : rnavigate.data.Sequence
            Sequence object to plot around the circle.
        structure : rnavigate.data.Structure, defaults to None
            Structure object to plot as lines within the circle.
        structure2 : rnavigate.data.Structure, defaults to None
            Structure object to compare to `structure`.
        interactions : rnavigate.data.Interactions, defaults to None
            Interactions object to plot as lines within the circle.
        interactions2 : rnavigate.data.Interactions, defaults to None
            Interactions object to plot as lines within the circle.
        profile : rnavigate.data.Profile, defaults to None
            Profile object used to color the sequence.
        annotations : list of rnavigate.data.Annotation, defaults to None
            List of Annotation objects to highlight regions around the circle.
        label : str, defaults to None
            Label for the plot title.
        colors : dict, defaults to None
            Dictionary of colors for each plot element.
        gap : float, defaults to 30
            Gap between the start and end of the sequence in degrees.
        nt_ticks : tuple of int, defaults to (20, 5)
            Gap between major and minor nucleotide ticks in degrees.
        """
        if annotations is None:
            annotations = []
        seq_circle = data.SequenceCircle(sequence, gap=gap)
        ax = self.get_ax()
        if colors is None:
            colors = {}
        colors = {
            "sequence": None,
            "nucleotides": "sequence",
        } | colors
        for key in ["nucleotides", "sequence"]:
            if isinstance(colors[key], str) and colors[key] == "contrast":
                continue
            elif colors[key] is None:
                continue
            colors[key], colormap = seq_circle.get_colors(
                colors[key],
                profile=profile,
                structure=structure,
                annotations=annotations,
            )
            self.add_colorbar_args(colormap)
        if isinstance(colors["sequence"], np.ndarray):
            colors["nucleotides"], _ = sequence.get_colors("white")
        elif colors["sequence"] == "contrast":
            colors["sequence"] = plots.get_contrasting_colors(colors["nucleotides"])
        if structure is not None:
            structure = structure.as_interactions(structure2)
        if colors["sequence"] is not None:
            plots.plot_sequence_ss(ax, seq_circle, colors["sequence"])
        if colors["nucleotides"] is not None:
            plots.plot_nucleotides_ss(ax, seq_circle, colors["nucleotides"])
        if structure is not None:
            plots.plot_interactions_circle(ax, seq_circle, structure)
            self.add_colorbar_args(structure.cmap)
        if interactions is not None:
            plots.plot_interactions_circle(ax, seq_circle, interactions)
            self.add_colorbar_args(interactions.cmap)
        if interactions2 is not None:
            plots.plot_interactions_circle(ax, seq_circle, interactions2)
            self.add_colorbar_args(interactions2.cmap)
        for annotation in annotations:
            plots.plot_annotation_circle(ax, seq_circle, annotation)
        self.set_axis(
            ax=ax, label=label, seq_circle=seq_circle, gap=gap, nt_ticks=nt_ticks
        )

    def set_axis(self, ax, label, seq_circle, gap, nt_ticks):
        """Set axis limits and ticks.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to modify.
        label : str
            Label for the plot title.
        seq_circle : rnavigate.data.SequenceCircle
            SequenceCircle object.
        gap : float
            Gap between the start and end of the sequence in degrees.
        nt_ticks : tuple of 2 integers
            Gap between major and minor nucleotide ticks in degrees.
        """
        ax.set_title(label)
        ticks, labels = plots.get_nt_ticks(
            sequence=seq_circle.sequence, region=(1, seq_circle.length), gap=nt_ticks[0]
        )
        tick_idx = seq_circle.data["Nucleotide"].isin(ticks)
        ticks = seq_circle.data.loc[tick_idx, "Theta"]
        ax.set_xticks(ticks)
        ax.set_xticklabels(labels)
        minor_ticks, _ = plots.get_nt_ticks(
            sequence=seq_circle.sequence,
            region=(1, seq_circle.length),
            gap=nt_ticks[1],
        )
        tick_idx = seq_circle.data["Nucleotide"].isin(minor_ticks)
        minor_ticks = seq_circle.data.loc[tick_idx, "Theta"]
        ax.set_xticks(minor=True, ticks=minor_ticks)
        ax.set(yticks=[], theta_zero_location="N", theta_direction=-1)
        ax.spines["polar"].set_bounds(np.pi * gap / 360, np.pi * (2 - gap / 360))
        # ax.spines["polar"].set_visible(False)
        ax.grid(False)
        self.i += 1



"""Provide the primary functions."""


def canvas(with_attribution=True):
    """
    Placeholder function to show example docstring (NumPy format).

    Replace this function and doc string for your own project.

    Parameters
    ----------
    with_attribution : bool, Optional, default: True
        Set whether or not to display who the quote is from.

    Returns
    -------
    quote : str
        Compiled string including quote and optional attribution.
    """

    quote = "The code is but a canvas to our imagination."
    if with_attribution:
        quote += "\n\t- Adapted from Henry David Thoreau"
    return quote


if __name__ == "__main__":
    # Do something if this file is invoked on its own
    print(canvas())



"""This module contains example rnav.Samples for different RNAs and data types.

TPP riboswitch in-vitro construct DMS-MaP data:
    All sequences are based on PDB entry 3DHS.
    Except for the PDB file, all other data also contains structure cassettes.
    Data keywords:
        pdb - 2GDI from the PDB
        ss - structure drawing diagram
        dmsmap - DMS-MaP structure probing data
        ringmap - RING-MaP correlations from DMS-MaP experiment
        pairprob - pairing probabilities using DMS-MaP as restraints

RNase P in-vitro construct SHAPE-MaP data
    There are five samples here:
    common reference data shared between all RNase P samples
    variable name (rnasep_common)
    data keywords:
        pdb - 3DHS, but with missing loops modelled in.
        shapejump - SHAPE-JuMP data
        ss_ct - 3DHS + cassettes, base-pairing structure with singlets removed
        ss_pdb - 3DHS, full base-pairing structure diagram
        pairprob - 3DHS + cassettes, pairing probabilities (sequence only)
    4 experimental samples: 2 SHAPE reagents * 2 folding conditions
    variable names (rnasep_1, rnasep_2, rnasep_3, rnasep_4)
    data keywords:
        all rnasep_common data keywords above +
        shapemap -  3DHS + cassettes, SHAPE-MaP experimental data
        ringmap - RING-MaP analysis on SHAPE-MaP data
        pairmap - PAIR-MaP analysis on SHAPE-MaP data

Example usage:

import rnavigate as rnav
from rnavigate.examples import rnasep_1, rnasep_2

plot = rnav.plot_arcs(
    samples=[rnasep_1, rnasep_2],
    sequence="ss_pdb",
    structure="ss_pdb",
    profile="shapemap",
    interactions="ringmap",
    )
"""
import rnavigate as rnav
from importlib import resources
from rnavigate.examples import (
    tpp_data,
    rmrp_data,
    rnasep_data,
    rrna_fragmap_data,
)


_rnasep_common = None
_rrna_fragmap_common = None


def __getattr__(name):  # pylint: disable=invalid-name
    if name == "tpp":
        tpp_dir = resources.files(tpp_data)
        return rnav.Sample(
            sample="TPP in-vitro DMS-MaP",
            pdb={"pdb": str(tpp_dir / "2gdi.pdb"), "chain": "X"},
            ss=str(tpp_dir / "TPP-2GDI.nsd"),
            dmsmap=str(tpp_dir / "DMS_TPP_profile.txt"),
            ringmap=str(tpp_dir / "DMS_TPP_rings.txt"),
            pairprob=str(tpp_dir / "TPP-dms-bp.dp"),
        )
    if name == "rmrp":
        rmrp_dir = resources.files(rmrp_data)
        return rnav.Sample(
            sample="RMRP in-vitro DMS-MaP",
            rnpmap=str(rmrp_dir / "RMRP-RNPMaP-Example_RESULTS.csv"),
            ss=str(rmrp_dir / "hs-RMRP.nsd"),
        )
    if name in ["rnasep_1", "rnasep_2", "rnasep_3", "rnasep_4"]:
        global _rnasep_common
        rnasep_dir = resources.files(rnasep_data)
        if _rnasep_common is None:
            _rnasep_common = rnav.Sample(
                sample="common data",
                pdb={
                    "pdb": str(rnasep_dir / "3dhsCrystal_PlusLoops.pdb"),
                    "chain": "A",
                },
                shapejump={
                    "shapejump": str(rnasep_dir / "example-rnasep-deletions.txt"),
                    "sequence": str(rnasep_dir / "RNaseP-noSC.fasta"),
                },
                ss_ct={"ss": str(rnasep_dir / "RNaseP.ct")},
                ss_pdb={"ss": str(rnasep_dir / "RC_CRYSTAL_STRUCTURE.xrna")},
                ss_lit={"ss": str(rnasep_dir / "RNaseP-lit-like.nsd")},
                pairprob={
                    "pairprob": str(rnasep_dir / "rnasep.dp"),
                    "sequence": str(rnasep_dir / "RNaseP-withSC.fasta"),
                },
            )
    if name == "rnasep_1":
        return rnav.Sample(
            sample="Example 1",
            inherit=_rnasep_common,
            shapemap={
                "shapemap": str(rnasep_dir / "example1_rnasep_profile.txt"),
                "log": str(rnasep_dir / "example1_shapemapper_log.txt"),
            },
            pairmap=str(rnasep_dir / "example1-rnasep-pairmap.txt"),
            ringmap=str(rnasep_dir / "example1-rnasep.corrs"),
        )
    if name == "rnasep_2":
        return rnav.Sample(
            sample="Example 2",
            inherit=_rnasep_common,
            shapemap={
                "shapemap": str(rnasep_dir / "example2_rnasep_profile.txt"),
                "log": str(rnasep_dir / "example2_shapemapper_log.txt"),
            },
            pairmap=str(rnasep_dir / "example2-rnasep-pairmap.txt"),
            ringmap=str(rnasep_dir / "example2-rnasep.corrs"),
        )
    if name == "rnasep_3":
        return rnav.Sample(
            sample="Example 3",
            inherit=_rnasep_common,
            shapemap={
                "shapemap": str(rnasep_dir / "example3_rnasep_profile.txt"),
                "log": str(rnasep_dir / "example3_shapemapper_log.txt"),
            },
            pairmap=str(rnasep_dir / "example3-rnasep-pairmap.txt"),
            ringmap=str(rnasep_dir / "example3-rnasep.corrs"),
        )
    if name == "rnasep_4":
        return rnav.Sample(
            sample="Example 4",
            inherit=_rnasep_common,
            shapemap={
                "shapemap": str(rnasep_dir / "example4_rnasep_profile.txt"),
                "log": str(rnasep_dir / "example4_shapemapper_log.txt"),
            },
            pairmap=str(rnasep_dir / "example4-rnasep-pairmap.txt"),
            ringmap=str(rnasep_dir / "example4-rnasep.corrs"),
        )
    if name in ["linezolid", "quinoxoline", "methyl"]:
        global _rrna_fragmap_common
        rrna_fragmap_dir = resources.files(rrna_fragmap_data)
        if _rrna_fragmap_common is None:
            _rrna_fragmap_common = rnav.Sample(
                sample="6HA1 - LSU",
                sequence=str(rrna_fragmap_dir / "6HA1_LSU.fasta"),
                pdb={
                    "pdb": str(rrna_fragmap_dir / "6HA1_LSU.pdb"),
                    "sequence": "sequence",
                    "chain": "A",
                },
                ss=str(rrna_fragmap_dir / "6HA1_LSU.json"),
            )
    if name == "quinaxoline":
        return rnav.Sample(
            sample="rRNA quinaxoline",
            inherit=_rrna_fragmap_common,
            shapemap=str(rrna_fragmap_dir / "2_QN_DMSO_subtracted_LSU_profile.txt"),
        )
    if name == "linezolid":
        return rnav.Sample(
            sample="rRNA linezolid",
            inherit=_rrna_fragmap_common,
            shapemap=str(rrna_fragmap_dir / "2_ZLD_DMSO_subtracted_LSU_profile.txt"),
        )
    if name == "methyl":
        return rnav.Sample(
            sample="rRNA methyl",
            inherit=_rrna_fragmap_common,
            shapemap=str(rrna_fragmap_dir / "2_Methyl_DMSO_subtracted_LSU_profile.txt"),
        )
    try:
        return globals()[name]
    except KeyError:
        raise AttributeError  # pylint: disable=raise-missing-from



"""Windowed AUROC assesses agreement between reactivities and base-pairing."""
from rnavigate import plots
from sklearn.metrics import roc_curve, auc
import numpy as np


class WindowedAUROC:
    """Compute and display windowed AUROC analysis.

    This analysis computes the ROC curve over a sliding window for the
    performance of per-nucleotide data (usually SHAPE-MaP or DMS-MaP Normalized
    reactivity) in predicting the base-pairing status of each nucleotide. The
    area under this curve (AUROC) is displayed compared to the median across
    the RNA. Below, an arc plot displays the secondary structure and
    per-nucleotide profile.

     AUROC values (should) range from 0.5 (no predictive power) to 1.0
    (perfect predictive power). A value of 0.5 indicates that the reactivity
    profile does not fit the structure prediction well. These regions are good
    candidates for further investigation with ensemble deconvolution.

    Citation:
    Lan, T.C.T., Allan, M.F., Malsick, L.E. et al. Secondary structural
        ensembles of the SARS-CoV-2 RNA genome in infected cells. Nat Commun
        13, 1128 (2022). https://doi.org/10.1038/s41467-022-28603-2

    Methods:
        __init__: Computes the AUROC array and AUROC median.
        plot_auroc: Displays the AUROC analysis over the given region.
            Returns Plot object

    Attributes:
        sample: an rnavigate.Sample to retrieve profile and secondary structure
        structure: sample.data[structure]
        profile: sample.data[profile]
        sequence: the sequence string of sample.data[structure]
        window: the size of the windows
        nt_length: the length of sequence string
        auroc: the auroc numpy array, length = nt_length, padded with np.nan
        median_auroc: the median of the auroc array
    """

    def __init__(
        self,
        sample,
        window=81,
        profile="default_profile",
        structure="default_structure",
    ):
        """Compute the AUROC for all windows. AUROC is a measure of how well a
        reactivity profile predicts paired vs. unpaired nucleotide status.

        Args:
            sample (rnav.Sample): Your rnavigate sample
            window (int, optional): number of nucleotides to include in window
                Defaults to 81.
            profile (str, optional): data keyword of provided sample pointing
                to a profile.
                Defaults to "default_profile"
            structure (str, optional): data keyword of provided sample pointing
                to a secondary structure.
                Defaults to "default_structure"
        """
        # ensure sample contains profile and structure data
        for data in [profile, structure]:
            assert data in sample.data.keys(), f"Sample missing {data} data"

        # store basic information
        self.sample = sample
        self.structure = sample.get_data(structure)
        self.profile = sample.get_data(profile)
        self.sequence = self.structure.sequence
        self.window = window
        self.nt_length = self.structure.length

        # get Norm_profile array and structure array
        profile = self.profile.data["Norm_profile"].values
        pair_nts = self.structure.pair_nts

        # for each possible window: compute auroc and populate array
        self.auroc = np.full(len(profile), np.nan)
        pad = window // 2
        for i in range(pad, len(profile) - pad):
            # get profile and structure values within window
            win_profile = profile[i - pad : i + pad + 1]
            win_ct = pair_nts[i - pad : i + pad + 1]
            # ignore positions where profile is nan
            valid = ~np.isnan(win_profile)
            # y: classification (paired or unpaired)
            y = win_ct[valid] == 0
            scores = win_profile[valid]
            # skip this window if there are less than 10 paired or unpaired nts
            if (sum(y) < 10) or (sum(~y) < 10):
                continue
            # add window auroc to array
            tpr, fpr, _ = roc_curve(y, scores)
            self.auroc[i] = auc(tpr, fpr)

        self.auroc_median = np.nanmedian(self.auroc)

    def plot_auroc(self, region=None):
        """Plot the result of the windowed AUROC analysis, with arc plot of
        structure and reactivity profile.

        Args:
            region (list of int: length 2, optional): Start and end nucleotide
                positions to plot. Defaults to [1, RNA length].
        """
        if region is None:
            start = 1
            stop = self.nt_length
            region = [start, stop]
            region_length = self.nt_length
        else:
            start, stop = region
            region_length = stop - start + 1

        plot = plots.AP(1, region_length, cols=1, rows=1, region=region)
        ax = plot.axes[0, 0]

        # fill between auroc values and median, using secondary ax
        x_values = np.arange(start, stop + 1)
        auc_ax = ax.twinx()
        auc_ax.set_ylim(0.5, 1.6)
        auc_ax.set_yticks([0.5, self.auroc_median, 1.0])
        auc_ax.fill_between(
            x_values,
            self.auroc[start - 1 : stop],
            self.auroc_median,
            fc="0.3",
            lw=0,
        )
        plots.adjust_spines(auc_ax, ["left"])
        plots.clip_spines(auc_ax, ["left"])

        # add structure and reactivity profile track
        plot.plot_data(
            sequence=self.structure,
            structure=self.structure,
            structure2=None,
            interactions=None,
            interactions2=None,
            profile=self.profile,
            label="label",
            seqbar=False,
            title=False,
            annotations=[],
            plot_error=False,
        )

        # Place Track Labels
        ax.set_title(
            f"{self.sample.sample}\n{start} - {stop}",
            loc="left",
            fontdict={"fontsize": 48},
        )
        ax.text(
            1.002,
            6 / 8,
            "Secondary\nStructure",
            transform=ax.transAxes,
            fontsize=36,
            va="center",
        )
        ax.text(
            1.002,
            2 / 8,
            f"{self.window}-nt window\nAUROC",
            transform=ax.transAxes,
            va="center",
            fontsize=36,
        )

        # limits, ticks, spines, and grid
        ax.set_ylim([-305, 315])
        ax.set_xticks(ticks=[x for x in range(500, stop, 500) if x > start])
        ax.set_xticks(ticks=[x for x in range(100, stop, 100) if x > start], minor=True)
        ax.tick_params(ax="x", which="major", labelsize=36)
        plots.adjust_spines(ax, ["bottom"])
        ax.grid(ax="x")

        # set figure size so that 100 ax units == 1 inch
        plot.set_figure_size(height_ax_rel=1 / 100, width_ax_rel=1 / 100)
        return plot
